"""This module is a command line script which lists the reference and/or
mapping files associated with the specified contexts by consulting the CRDS
server.   More generally it's for printing out information on CRDS files.
"""
import os.path
import sys
from collections import OrderedDict
import json

# ============================================================================

# from astropy.io import fits   # deferred

# ============================================================================

import crds
from crds.core import config, log, rmap, heavy_client, cmdline, utils
from crds.core import crds_cache_locking
from crds.core import reftypes

from crds import data_file

from crds.client import api

# ============================================================================

class ListScript(cmdline.ContextsScript):
    """Command line script for listing a variety of information about CRDS."""

    description = """crds.list prints out a variety of information about CRDS configuration, the
cache, reference and mapping files, default context names, and dataset headers
and ids used for CRDS reprocessing recommendations.
    """

    epilog = """
General categories of information driven by switches include:

0. Overall CRDS configuration
1. CRDS server file lists
2. CRDS cache file lists and paths
3. Cached file contents or headers
4. CRDS reprocessing dataset ids and parameters
5. Listing global default and installed pipeline contexts
6. Resolving context specifiers into literal context names

Many crds list services require setting CRDS_SERVER_URL to a valid CRDS
server to provide a source for the headers.

For HST::

% export CRDS_SERVER_URL=https://hst-crds.stsci.edu

or for JWST::

% export CRDS_SERVER_URL=https://jwst-crds.stsci.edu

0. Configuration information governing the behavior of CRDS for simple
configurations can be dumped::

crds list --status
CRDS Version = '7.0.7, bump-version, 7432326'
CRDS_MODE = 'auto'
CRDS_PATH = '/Users/jmiller/crds_cache_ops'
CRDS_SERVER_URL = 'https://jwst-crds.stsci.edu'
Effective Context = 'jwst_0204.pmap'
Last Synced = '2016-09-20 08:00:09.115330'
Python Executable = '/Users/jmiller/anaconda/bin/python'
Python Version = '3.5.2.final.0'
Readonly Cache = False

More comprehensive configuration information is also available for advanced
configurations::

% crds list --config
... lots of info ....

1. Files known by the CRDS server to belong to specified contexts can be listed
even if the files are not installed in a local CRDS Cache.

The --mappings command recursively evaluates and includes all the sub-mappings,
i.e. imaps and pmaps, of the specified contexts.

Contexts to list can be specified in a variety of ways:

-- To list the references contained by several contexts::

% crds list  --references --contexts hst_0001.pmap hst_0002.pmap ...
vb41935ij_bia.fits
vb41935kj_bia.fits
...

-- To list the references in a numerical range of contexts::

% crds list --references --range 1:2 --references
vb41935lj_bia.fits
vb41935oj_bia.fits
...

-- To list all mappings, even those not referenced by an imap or pmap::

% crds list --mappings --all
hst.pmap
hst_0001.pmap
hst_0002.pmap
hst_acs.imap
hst_acs_0001.imap
hst_acs_0002.imap
hst_acs_atodtab.rmap
...

--references, --mappings, or both can be listed.

2. Locally cached files (files already synced to your computer) can be listed::

% crds list --cached-mappings --full-path
...

% crds list --cached-references --full-path
...

In both cases adding --full-path prints the path of the file within the CRDS cache.

These are merely simple directory listings which ignore the context specifiers
and can be grep'ed for finer grained answers.

3. The contents of cached mappings or references (header only) can be printed to stdout like this::

% crds list --contexts jwst-fgs-linearity-edit jwst-nirspec-linearity-edit --cat --add-filename | grep parkey
CRDS - INFO - Symbolic context 'jwst-fgs-linearity-edit' resolves to 'jwst_fgs_linearity_0008.rmap'
CRDS - INFO - Symbolic context 'jwst-nirspec-linearity-edit' resolves to 'jwst_nirspec_linearity_0009.rmap'
/cache/path/mappings/jwst/jwst_fgs_linearity_0008.rmap:     'parkey' : (('META.INSTRUMENT.DETECTOR', 'META.SUBARRAY.NAME'), ('META.OBSERVATION.DATE', 'META.OBSERVATION.TIME')),
/cache/path/mappings/jwst/jwst_nirspec_linearity_0009.rmap:     'parkey' : (('META.INSTRUMENT.DETECTOR', 'META.SUBARRAY.NAME'), ('META.OBSERVATION.DATE', 'META.OBSERVATION.TIME')),

this prints the contents of the specified rmaps.

The -edit specifier above refers to mappings contained by the default starting point (.pmap) of future
server submissions.  It tracks on-going submission work that precedes the adoption of a new context
as the default in use by the pipeline.

crds.list --cat can be applied to references and prints out the reference metadata that CRDS views
abstractly as the file header.

References need to be catted explicitly by name,  but the list can come from the --references command
explained above::

% crds list --cat jwst_nirspec_dark_0036.fits
CRDS - INFO - Symbolic context 'jwst-operational' resolves to 'jwst_0167.pmap'
##########################################################################################
File:  '/grp/crds/jwst/references/jwst/jwst_nirspec_dark_0036.fits'
##########################################################################################
{'A1_COL_C': '8.9600000e+002',
'A1_CONF1': '2.1846000e+004',
...
}

4. Information about the dataset IDs and parameters used for CRDS reprocessing
and regressions can be printed or stored.

 Parameter set IDs can be listed for one or more instruments as follows::

 % crds list --dataset-ids-for-instruments wfc3...
 JCL403010:JCL403ECQ
 ... hundreds to hundreds of thousands of IDs as shown above ...

 IDs can also be captured to a file using UNIX I/O redirection::

 % crds list --dataset-ids-for-instruments wfc3   >wfc3.ids

 IDs for HST are of the form <product>:<exposure> where many exposures feed into
 the construction of one product and recalibrating any component exposure suggests
 recalibrating the combined product.

 CRDS stores dataset parameters for regression testing as a JSON dictionaries
 specifying one set of dataset parameters per line of the file::

 % crds list --dataset-headers @wfc3.ids --json > wfc3.headers.json

 NOTE:  while IDs can be specified directly on the command line,  CRDS has an
 @-notation that means "take IDs from this file".

 The JSON headers are suitable for running through crds.bestrefs to perform
 reprocessing checks or single context reference file coverage checks shown  here::

 % crds bestrefs --load-pickle wfc3.headers.json --dump-unique-errors --stats
 ...  errors related to looking up references for these parameter sets ...

 The script crds_dataset_capture combines the process of dumping all IDs for an
 instrument and dumping their corresponding dataset parameters.  IDs files and
 header files are placed in a dated regression capture directory::

 % crds_dataset_capture wfc3 acs ...
 ... downloads IDs and headers for WFC3, ACS to dated directory ...

 The default multi-line format for dataset parameters is more readable than the
 --json form::

 % crds list --dataset-headers jcl403010 --first-id --minimize-header
 CRDS - INFO - Symbolic context 'hst-operational' resolves to 'hst_0462.pmap'
 CRDS - INFO - Dataset pars for 'JCL403010:JCL403ECQ' with respect to 'hst_0462.pmap'
 {'APERTURE': 'WFC1',
  'ATODCORR': 'OMIT',
  'BIASCORR': 'COMPLETE',
  'CCDAMP': 'ABCD',
  'CCDCHIP': '-999.0',
  'CCDGAIN': '2.0',
  'CRCORR': 'OMIT',
  'DARKCORR': 'COMPLETE',
  'DATE-OBS': '2016-02-20',
  'DETECTOR': 'WFC',
  'DQICORR': 'COMPLETE',
  'DRIZCORR': 'COMPLETE',
  'FILTER1': 'CLEAR1L',
  'FILTER2': 'F814W',
  'FLASHCUR': 'LOW',
  'FLATCORR': 'COMPLETE',
  'FLSHCORR': 'OMIT',
  'FW1OFFST': '0.0',
  'FW2OFFST': '0.0',
  'FWSOFFST': '0.0',
  'GLINCORR': 'UNDEFINED',
  'INSTRUME': 'ACS',
  'LTV1': '0.0',
  'LTV2': '0.0',
  'NAXIS1': '4144.0',
  'NAXIS2': '4136.0',
  'OBSTYPE': 'IMAGING',
  'PCTECORR': 'UNDEFINED',
  'PHOTCORR': 'COMPLETE',
  'RPTCORR': 'UNDEFINED',
  'SHADCORR': 'OMIT',
  'SHUTRPOS': 'A',
  'TIME-OBS': '17:32:29.666665',
  'XCORNER': '0.0',
  'YCORNER': '0.0',
  'dataset_id': 'JCL403010:JCL403ECQ'}

 Sometimes it's desirable to know the individual exposures CRDS associates with a product id::

 % crds list --dataset-headers jcl403010 --id-expansions-only
 CRDS - INFO - Symbolic context 'hst-operational' resolves to 'hst_0462.pmap'
 JCL403010:JCL403ECQ
 JCL403010:JCL403EEQ
 JCL403010:JCL403EGQ
 JCL403010:JCL403EIQ
 JCL403010:JCL403EKQ
 JCL403010:JCL403EMQ
 JCL403010:JCL403EOQ
 JCL403010:JCL403EQQ
 JCL403010:JCL403ESQ
 JCL403010:JCL403EUQ

5. Information about the default context can be printed.  There are two variations and a subtle distinction::

% python m crds.list --operational-context
jwst_0204.pmap

lists the context which has been *commanded* as default on the CRDS server.

While::

% crds list --remote-context jwst-ops-pipeline
jwst_0101.pmap

lists the context which is *in actual use* in the associated archive pipeline as reported by
a cache sync echo.

During the interval between commanding a new default on the CRDS server and syncing the pipeline
CRDS cache,  the commanded and actual pipeline contexts can differ.

6. Resolving context specifiers

Some CRDS tools, including crds.list and crds.sync, support multiple
mechanisms for specifying context.  The --resolve-contexts command
interprets those specifiers into a non-recursive list of literal mapping
names and prints them out.  --resolve-contexts differs from --mappings
because it does not implicitly include all sub-mappings of the specified
contexts::

% crds list --resolve-contexts --all
jwst.pmap
jwst_0000.pmap
jwst_0001.pmap
jwst_0002.pmap
jwst_0003.pmap
...

% crds list --resolve-contexts --last 5
jwst_0205.pmap
jwst_0206.pmap
jwst_0207.pmap
jwst_0208.pmap
jwst_0209.pmap

% crds list --resolve-contexts  --contexts jwst-miri-dark-operational
jwst_miri_dark_0012.rmap

% crds list --resolve-contexts --contexts jwst-niriss-superbias-2016-01-01T00:00:00
jwst_niriss_superbias_0005.rmap
    """
    def __init__(self, *args, **keys):
        super(ListScript, self).__init__(*args, **keys)
        self.show_context_resolution = False
        self._file_info = {}

    def add_args(self):
        """Add switches unique to crds.list."""

        self.add_argument('--references', action='store_true', dest="list_references",
            help='print names of reference files referred to by contexts')
        self.add_argument('--mappings', action='store_true', dest="list_mappings",
            help='print names of mapping files referred to by contexts')
        self.add_argument("--cached-references", action="store_true",
            help="print the paths of all references in the local cache. very primitive.")
        self.add_argument("--cached-mappings", action="store_true",
            help="print the paths of all mappings in the local cache. very primitive.")
        self.add_argument("--cached-pickles", action="store_true",
            help="print the paths of all pickles in the local cache. very primitive.")
        self.add_argument("--full-path", action="store_true",
            help="print the full paths of listed files.")

        self.add_argument("--dataset-ids-for-instruments", nargs="+", dest="dataset_ids", default=None, metavar="INSTRUMENTS",
            help="print the dataset ids known to CRDS associated for the specified instruments.")
        self.add_argument("--dataset-headers", nargs="+", dest="dataset_headers", default=None, metavar="IDS",
            help="print matching parameters for the specified dataset ids.")
        self.add_argument("--id-expansions-only", action="store_true", dest="id_expansions_only",
            help="print out only the <product>:<exposure> expansion associated with the specified --dataset-headers ids.")
        self.add_argument("--first-id-expansion-only", action="store_true", dest="first_id_expansion_only",
            help="print out only the first exposure ID (header or expanded) associated with a particular product ID.")
        self.add_argument("--minimize-headers", action="store_true", dest="minimize_headers",
            help="print out only header parameters required by a particular CRDS context.")
        self.add_argument("--json", action="store_true", dest="json",
            help="print out header parameters in JSON format suited for crds.bestrefs and grepping.")

        self.add_argument("--config", action="store_true", dest="config",
            help="print CRDS configuration information.")

        self.add_argument("--status", action="store_true", dest="status",
            help="print brief, basic, CRDS configuration information.")

        self.add_argument("--cat", nargs="*", dest="cat", metavar="FILES", default=None,
            help="print the text of the specified mapping files or header and format info for references.")
        self.add_argument("--keywords", nargs="+",
            help="limited list of keywords to be catted from reference headers.")
        self.add_argument("--add-filenames", action="store_true",
            help="prefix each line of a cat'ed file with the filename.")
        self.add_argument("--no-arrays", action="store_true",
            help="Don't --cat array properties that are slow to compute. Use for large files.")

        self.add_argument("--tpns", nargs="*", dest="tpns", metavar="FILES", default=None,
            help="print the certify constraints (.tpn's) associated with the specified or implied files.")

        self.add_argument("--operational-context", action="store_true", dest="operational_context",
            help="print the name of the operational context on the central CRDS server.")
        self.add_argument("--remote-context", type=str, metavar="PIPELINE",
            help="print the name of the context reported as in use by the specified pipeline.")
        self.add_argument("--resolve-contexts", action="store_true", dest="resolve_contexts",
            help="print the literal names of the contexts defined by the command line context specifiers.")

        self.add_argument("--required-parkeys", action="store_true",
            help="print the names of the parkeys required to compute bestrefs for the specified mappings.")

        self.add_argument("--file-properties", nargs="*", default=None,
                          help="print the instrument, filekind, filename for each of the files specified.")

        super(ListScript, self).add_args()

    def main(self):
        """List files."""
        if self.args.cat is not None: # including []
            return self.cat_files()

        if self.args.tpns:
            return self.list_tpns()

        if self.args.file_properties is not None: # including []
            return self.list_file_properties()

        if self.args.operational_context:
            print(self.default_context)
            return
        if self.args.remote_context:
            print(self.remote_context)
            return

        if self.args.resolve_contexts:
            self.list_resolved_contexts()

        if self.args.list_references:
            self.list_references()
        if self.args.list_mappings:
            self.list_mappings()
        if self.args.cached_references:
            self.list_cached_references()
        if self.args.cached_mappings:
            self.list_cached_mappings()
        if self.args.cached_pickles:
            self.list_cached_pickles()

        if self.args.dataset_ids:
            self.list_dataset_ids()
        if self.args.dataset_headers:
            self.list_dataset_headers()

        if self.args.config:
            self.list_config()
        if self.args.status:
            self.list_status()

        if self.args.required_parkeys:
            self.list_required_parkeys()

        return log.errors()

    def list_resolved_contexts(self):
        """Print out the literal interpretation of the contexts implied by the script's
        context specifiers.
        """
        self.show_context_resolution = True
        for context in self.contexts:
            print(context)

    @property
    def remote_context(self):
        """Print the name of the context in use at pipeline `self.args.remote_context`
        as recorded on the server after being pushed by the crds.sync tool in the pipeline.
        """
        self.require_server_connection()
        with log.error_on_exception("Failed resolving remote context"):
            return api.get_remote_context(self.observatory, self.args.remote_context)

    @property
    def implied_files(self):
        """Return the filenames implied by a combination of the specified contexts
        and whether or not --mappings or --references are set.
        """
        mappings = self.get_context_mappings() if self.args.list_mappings else []
        references = self.get_context_references() if self.args.list_references else []
        return [filename for filename in mappings + references if filename != 'N/A']

    def list_file_properties(self):
        """Print out the (instrument, filekind, filename) information for each of the
        files implied by --files, and any contexts plus --mappings and/or --references.
        """
        for filename in self.args.file_properties + self.implied_files:
            with log.error_on_exception("Failed dumping file properties for", repr(filename)):
                filepath = self.locate_file(filename)
                instrument, filekind = self.get_file_properties(filepath)
                print(instrument, filekind, os.path.basename(filepath))

    def cat_files(self):
        """Print out the files listed after --cat or implied by a combination of
        explicitly specified contexts and --mappings or --references.

        --files is not allowed.

        """
        # --cat files...   specifying *no* files still triggers --cat logic
        # --contexts context-specifiers [including --all --last --range...]
        # context specifiers can be symbolic and will be resolved.
        # --cat @file is allowed

        catted_files = self.get_words(self.args.cat) + self.implied_files
        try:
            self._file_info = api.get_file_info_map(
                self.observatory, files=[os.path.basename(filename) for filename in catted_files])
        except Exception:
            log.verbose_warning("Failed retrieving CRDS server catalog information.  May need to set CRDS_SERVER_URL.")

        # This could be expanded to include the closure of mappings or references
        for name in catted_files:
            with log.error_on_exception("Failed dumping:", repr(name)):
                path = self.locate_file(name)
                self._cat_file(path)

    def _cat_file(self, path):
        """Print out information on a single reference or mapping at `path`."""
        self._cat_banner("File:", os.path.abspath(path), delim="#", bottom_delim="-")
        if config.is_reference(path):
            self._cat_reference(path)
        else:
            self._cat_text(path)
        if self._file_info:
            self._cat_catalog_info(path)

    def _cat_banner(self, *args, **keys):
        """Print a banner for --cat for `name` and return the filepath of `name`."""
        delim = keys.get("delim","#")
        bottom_delim = keys.get("bottom_delim", None)
        if delim:
            print(delim*80)
        print(*args)
        if bottom_delim:
            print(bottom_delim*80)

    def _cat_reference(self, path):
        """Print information on any reference type at `path`."""
        self._cat_header(path)
        if path.endswith(".fits"):
            self._cat_banner("Fits Info:", delim="-", bottom_delim=None)
            self._print_lines(path, _fits_info_lines(path))
            if not self.args.no_arrays:
                self._cat_array_properties(path)

    def _cat_array_properties(self, path):
        """Print out the CRDS interpretation of every array in `path`,  currently FITS only."""
        with data_file.fits_open(path) as hdulist:
            for i, hdu in enumerate(hdulist):
                with log.warn_on_exception("Can't load array properties for HDU[" + str(i) +"]"):
                    if i > 0:
                        extname = hdu.header.get("EXTNAME",str(i))
                        self._cat_banner("CRDS Array Info [" + repr(extname) + "]:",
                                         delim="-", bottom_delim=None)
                        props = data_file.get_array_properties(path, extname)
                        props = { prop:value for (prop,value) in props.items() if value is not None }
                        self._print_lines(path, _pp_lines(props))

    def _cat_text(self, path):
        """Dump out the contexts of a text file."""
        with open(path) as pfile:
            self._print_lines(path, pfile.readlines())

    def _cat_header(self, path):
        """Dump out the header associated with a reference file."""
        old = config.ALLOW_SCHEMA_VIOLATIONS.set(True)
        header = data_file.get_unconditioned_header(path)
        if self.args.keywords:
            header2 = {}
            for keyword in header:
                for substr in self.args.keywords:
                    if substr in keyword:
                        header2[keyword] = header[keyword]
        else:
            header2 = header
        self._print_lines(path, _pp_lines(header2))
        config.ALLOW_SCHEMA_VIOLATIONS.set(old)

    def _cat_catalog_info(self, path):
        """Dump out all the info contained in the CRDS server file catlog."""
        try:
            info = self._file_info[os.path.basename(path)]
            if isinstance(info, dict):
                info.pop("deliverer_user")
                self._cat_banner("Catalog Info:", delim="-", bottom_delim=".")
                self._print_lines(path, _pp_lines(info))
        except KeyError:
            print("Server catalog info for", repr(path), "not available.")

    def _print_lines(self, path, lines):
        """Print `lines` to stdout,  optionally prefixing each line with `path`."""
        for line in lines:
            if self.args.add_filenames:
                print(os.path.basename(path) + ":", line.rstrip())
            else:
                print(line.rstrip())

    def list_references(self):
        """Consult the server and print the names of all references associated with
        the given contexts.
        """
        references = [ config.locate_file(filename, self.observatory) if self.args.full_path else filename
                       for filename in self.get_context_references() ]
        _print_list(references)

    def list_mappings(self):
        """Consult the server and print the names of all CRDS mappings associated
        with the given contexts.
        """
        mappings = [ config.locate_file(filename, self.observatory) if self.args.full_path else filename
                       for filename in self.get_context_mappings() ]
        _print_list(mappings)

    def list_cached_mappings(self):
        """List the mapping paths in the local cache."""
        _print_list(rmap.list_mappings("*.*map", self.observatory, full_path=self.args.full_path))

    def list_cached_pickles(self):
        """List the pickle paths in the local cache."""
        _print_list(rmap.list_pickles("*", self.observatory, full_path=self.args.full_path))

    def list_cached_references(self):
        """List the reference paths in the local cache."""
        _print_list(rmap.list_references("*", self.observatory, full_path=self.args.full_path))

    def list_dataset_headers(self):
        """List dataset header info for self.args.dataset_headers with respect to self.args.contexts"""
        # Support @-files for ids specified on command line
        ids = self.get_words(self.args.dataset_headers)
        products_seen, exposures_seen = set(), set()
        expanded_ids = []
        for context in self.contexts:
            with log.error_on_exception("Failed fetching dataset parameters with repect to", repr(context),
                                        "for", repr(self.args.dataset_headers)):
                for returned_id, header in api.get_dataset_headers_unlimited(context, ids):
                    product, exposure = returned_id.split(":")
                    if isinstance(header, str):
                        log.error("No header for", repr(returned_id), ":", repr(header)) # header is reason
                        continue
                    if self.args.first_id_expansion_only and product in products_seen:
                        continue
                    products_seen.add(product)
                    exposures_seen.add(exposure)
                    if self.args.id_expansions_only:
                        expanded_ids += [ returned_id + (" " + context if len(self.contexts) > 1 else "")]
                    else:
                        self.dump_header(context, returned_id, header)
        if self.args.id_expansions_only:
            for expanded in sorted(expanded_ids):
                print(expanded)

    def dump_header(self, context, returned_id, header):
        """Print out dataset `header` for `id` and `context` in either .json or
        multi-line formats.
        """
        pmap = crds.get_cached_mapping(context)
        if self.args.minimize_headers:
            header2 = pmap.minimize_header(header)
        else:
            header2 = dict(header)
        header2.pop("REFTYPE", None)
        header2["dataset_id"] = returned_id
        header2["CRDS_CTX"] = context
        if self.args.json:
            json_header = { returned_id : header }
            print(json.dumps(json_header))
        else:
            print("Dataset pars for", repr(returned_id), "with respect to", repr(context) + ":\n",
                  log.PP(header2))

    def list_dataset_ids(self):
        """Print out the dataset ids associated with the instruments specified as command line params."""
        for instrument in self.args.dataset_ids:
            with log.error_on_exception("Failed reading dataset ids for", repr(instrument)):
                for context in self.contexts:
                    ids = api.get_dataset_ids(context, instrument)
                    for dataset_id in ids:
                        if len(self.contexts) > 1:
                            print(context, dataset_id)
                        else:
                            print(dataset_id)

    def list_config(self):
        """Print out configuration info about the current environment and server."""
        info = config.get_crds_env_vars()
        real_paths = config.get_crds_actual_paths(self.observatory)
        server = self.server_info
        current_server_url = api.get_crds_server()
        cache_subdir_mode = config.get_crds_ref_subdir_mode(self.observatory)
        pyinfo = _get_python_info()
        _print_dict("CRDS Environment", info)
        _print_dict("CRDS Client Config", {
                "server_url" : current_server_url,
                "cache_subdir_mode": cache_subdir_mode,
                "readonly_cache": self.readonly_cache,
                "effective_context": heavy_client.get_context_name(self.observatory),
                "crds" : repr(crds),
                "version": heavy_client.version_info()
                })
        _print_dict("CRDS Actual Paths", real_paths)
        _print_dict("CRDS Server Info", server,
                    ["observatory", "status", "connected", "operational_context", "last_synced",
                     "reference_url", "mapping_url", "effective_mode"])
        download_uris = dict(
            config_uri = os.path.dirname(api.get_flex_uri("server_config")),
            pickle_uri = os.path.dirname(api.get_flex_uri("xyz.pmap.pkl")),
            mapping_uri = os.path.dirname(api.get_flex_uri("xyz.pmap")),
            reference_uri = os.path.dirname(api.get_flex_uri("xyz.fits")),
            )
        _print_dict("Actual Download URIs", download_uris)
        if self.observatory == "hst":
            cal_vars = { var : os.environ[var] for var in os.environ
                          if len(var) == 4 and var.lower().endswith("ref") }
            _print_dict("Calibration Environment", cal_vars)
        _print_dict("Python Environment", pyinfo)

    def list_status(self):
        """Print out *basic* configuration info about the current environment and server."""
        info = config.get_crds_env_vars()
        server = self.server_info
        pyinfo = _get_python_info()
        status = OrderedDict(
            [("CRDS_PATH", info.get("CRDS_PATH", "undefined")),
             ("CRDS_SERVER_URL", info.get("CRDS_SERVER_URL", "undefined")),
             ("CRDS_MODE", info["CRDS_MODE"]),
             ("Readonly Cache", self.readonly_cache),
             ("Cache Locking", crds_cache_locking.status()),
             ("Effective Context", heavy_client.get_context_name(self.observatory)),
             ("Last Synced", server.last_synced),
             ("CRDS Version", heavy_client.version_info()),
             ("CRDS Summary", crds.__rationale__),
             ("Python Version", pyinfo["Python Version"]),
             ("Python Executable", pyinfo["Python Executable"]),
             ])
        _print_dict(None, status)

    def list_required_parkeys(self):
        """Print out the parkeys required for matching using the specified contexts."""
        for name in self.contexts:
            mapping = crds.get_cached_mapping(name)
            if isinstance(mapping, rmap.PipelineContext):
                log.divider(name="Parkeys required for " + repr(mapping.basename), func=log.write)
                _print_dict("", mapping.get_required_parkeys())
            elif isinstance(mapping, rmap.InstrumentContext):
                for name in sorted(mapping.selections):
                    try:
                        rmapping = mapping.get_rmap(name)
                    except (crds.exceptions.IrrelevantReferenceTypeError,
                            crds.exceptions.OmitReferenceTypeError):
                        print(name +":", repr("N/A"))
                    else:
                        print(name + ":",  rmapping.get_required_parkeys())
            else:
                print(name + ":",  mapping.get_required_parkeys())

    def list_tpns(self):
        """Print out the .tpn information related to each of the files either
        specified via the --tpns <files...> argument or implied by any of the
        other standard --list mechanisms like --mappings or --references.
        """
        constrained_files = self.get_words(self.args.tpns) + self.implied_files
        for filename in constrained_files:
            path = self.locate_file(filename)
            if config.is_mapping(path):
                tpn_text = reftypes.get_types_object(self.observatory).reference_name_to_ld_tpn_text(path)
            else:
                tpn_text = reftypes.get_types_object(self.observatory).reference_name_to_tpn_text(path)
            log.divider(f"Certify constraints for '{path}'")
            print(tpn_text)

def _get_python_info():
    """Collect and return information about the Python environment"""
    pyinfo = {
        "Python Version" : ".".join(str(num) for num in sys.version_info),
        "Python Executable": sys.executable,
        }
    pypath = os.environ.get("PYTHON_PATH", None)
    if pypath:
        pyinfo["PYTHON_PATH"] = pypath
    return pyinfo

def _print_dict(title, dictionary, selected = None):
    """Print out dictionary `d` with a one line `title`."""
    if selected is None:
        selected = dictionary.keys()
    if title:
        print(title)
    if dictionary:
        for key in sorted(selected):
            try:
                print(("\t" if title else "") + key + " = " + repr(dictionary[key]))
            except Exception:
                print(("\t" if title else "") + key + " = " + repr(getattr(dictionary, key)))
    else:
        print("\t" + "none")

def _print_list(files):
    """Print `files` one file per line."""
    for filename in files:
        print(filename)

def _fits_info_lines(path):
    """Return the output of fits.info() as a list of line strings"""
    return _captured_fits_info.outputs(path).splitlines()

@utils.capture_output
def _captured_fits_info(path):
    """Return fits info as a string."""
    from astropy.io import fits   # deferred
    fits.info(path)

def _pp_lines(obj):
    """Pretty print `obj` and return the resulting text as a list of lines."""
    return str(log.PP(obj)).splitlines()

if __name__ == "__main__":
    sys.exit(ListScript()())



import os
import shutil
import tempfile
from nose.tools import assert_equals, assert_not_equals, assert_is, assert_true, raises
from crds.submit import Submission, NoFilesSelected
from crds.tests import test_config
import mock

# To run:
#   nosetests -v unit_tests.py

TEMPFILES = ['ipppssoot_ccd.fits', 'opppssoot_bia.fits']

# Mocked urllib.request to .../redcat_description.yml:
FORM_DESCRIPTION_YML = '''\
- {help_text: 'Who are you?', key: deliverer, label: Name of deliverer, required: true,
  type: CharField}
- {help_text: Comma-delimited list (optional), key: other_email, label: Other e-mail
    adresses to send notifications, required: false, type: CharField}
- choices: [acs, cos, nicmos, stis, synphot, wfc3, wfpc2]
  initial: acs
  key: instrument
  label: Instrument  (All submitted files should match this instrument.  This instrument
    will be locked for your submission exclusively)
  required: true
  type: TypedChoiceField
- {key: file_type, label: 'Type of files (Bias, Dark, etc.)', required: true, type: CharField}
- choices: [false, true]
  initial: false
  key: history_updated
  label: Has HISTORY section in the primary header been updated to describe in detail
    the reason for delivery and how the files were created?
  required: false
  type: BooleanField
- choices: [false, true]
  initial: false
  key: pedigree_updated
  label: Has PEDIGREE keyword been checked and updated as necessary?
  required: false
  type: BooleanField
- choices: [false, true]
  initial: false
  key: keywords_checked
  label: Has COMMENT been checked?
  required: false
  type: BooleanField
- choices: [false, true]
  initial: false
  key: descrip_updated
  label: Was the DESCRIP keyword updated with a summary of why the files were updated
    or created?
  required: false
  type: BooleanField
- choices: [false, true]
  initial: false
  key: useafter_updated
  label: Has the USEAFTER keyword been checked, and if necessary, updated?
  required: false
  type: BooleanField
- choices: [N/A, 'No', 'Yes']
  help_text: N/A for ETC Files Only
  initial: N/A
  key: useafter_matches
  label: If the reference files are replacing previous versions, do the new USEAFTER
    dates exactly match the old ones?
  required: true
  type: TypedChoiceField
- choices: [N/A, 'No', 'Yes']
  help_text: optional
  initial: N/A
  key: compliance_verified
  label: Verification for compliance complete (fits, certify, etc. or N/A)
  required: true
  type: TypedChoiceField
- choices: [false, true]
  initial: false
  key: ingest_files
  label: Should the affected files be reprocessed?
  required: false
  type: BooleanField
- choices: [false, true]
  initial: false
  key: etc_delivery
  label: Should the files be submitted to the ETC?
  required: false
  type: BooleanField
- choices: [false, true]
  initial: false
  key: jwst_etc
  label: Are these JWST ETC files?
  required: false
  type: BooleanField
- {key: calpipe_version, label: Files run through the current version of the calibration
    software being used by the pipeline or PYSYNPHOT and ETC (yes/no and version number),
  required: true, type: CharField}
- choices: [false, true]
  initial: false
  key: replacement_files
  label: Are any files replacing old reference files (deliveries can be a mix of files
    that are or are not replacing old files) (yes/no)
  required: false
  type: BooleanField
- {key: old_reference_files, label: 'If yes, list them here', required: false, type: CharField}
- choices: [N/A, 'No', 'Yes']
  initial: N/A
  key: replacing_badfiles
  label: If the files being replaced are bad, and should not be used with any data,
    please indicate this here
  required: true
  type: TypedChoiceField
- {help_text: Comma-delimited list (optional), key: jira_issue, label: Any JIRA issues
    filed in regard to the references being delivered (e.g. "REDCAT-25"), required: false,
  type: CharField}
- {key: table_rows_changed, label: 'If files are tables, please indicate exactly which
    rows have changed', required: false, type: CharField}
- {key: modes_affected, label: 'Please indicate which modes (e.g. all the STIS, FUVMAMA,
    E140L modes) are affected by the changes in the files', required: true, type: CharField}
- {key: correctness_testing, label: Description of how the files were tested for correctness,
  required: true, type: CharField}
- {key: additional_considerations, label: Additional considerations, required: false,
  type: CharField}
'''

def touch(path):
    with open(path, 'a'):
        os.utime(path, None)

class TestSubmission(object):
    @classmethod
    @mock.patch('crds.submit.rc_submit.urllib.request.urlopen', autospec=True)
    def setup_class(cls, urlopen):
        '''This method is run once for each class before any tests are run.'''
        cls.old_state = test_config.setup()

        # Create a temporary directory:
        cls.tmpdir = tempfile.mkdtemp(prefix='tmp_rc_submit_')

        # Create empty test files in the temporary directory:
        cls.tempfiles = [os.path.join(cls.tmpdir, x) for x in TEMPFILES]
        for filename in cls.tempfiles:
            touch(filename)

        # Create a file handle to use as a mockup of the urllib.request object:
        cls.mockup_form = os.path.join(cls.tmpdir, 'mocked_redcat_description.yml')
        with open(cls.mockup_form, 'w') as f:
            f.write(FORM_DESCRIPTION_YML)
        urlopen.return_value = open(cls.mockup_form)

        # Instantiate the Submission object used in these tests:
        cls.s = Submission('hst', 'dev', context='hst_0723.pmap')

    @classmethod
    def teardown_class(cls):
        '''This method is run once for each class after all tests are run.'''
        # Remove temporary directory and all files contained therein:
        shutil.rmtree(cls.tmpdir)
        test_config.cleanup(cls.old_state)

    @raises(KeyError)
    def test_badkey(self):
        self.s['bad_key'] = 'some value'

    def test_goodvalue_char(self):
        self.s['file_type'] = 'bias'

    def test_goodvalue_bool(self, key='history_updated'):
        self.s[key] = True
        assert_is(self.s[key], True)
        self.s[key] = False
        assert_is(self.s[key], False)

    def test_goodvalue_trinary(self, key='compliance_verified'):
        # Set with Booleans:
        self.s[key] = True
        assert_equals(self.s[key], 'Yes')
        self.s[key] = False
        assert_equals(self.s[key], 'No')

        # Set with strings:
        self.s[key] = 'Yes'
        assert_equals(self.s[key], 'Yes')
        self.s[key] = 'yes'  # Handle different case
        assert_equals(self.s[key], 'Yes')
        self.s[key] = 'No'
        assert_equals(self.s[key], 'No')
        self.s[key] = 'n/a'  # Handle different case
        assert_equals(self.s[key], 'N/A')

    @raises(ValueError)
    def test_badtype(self):
        self.s['calpipe_version'] = 123  # Expects a str

    @raises(ValueError)
    def test_badvalue_trinary(self, key='compliance_verified'):
        self.s[key] = 'bad value'

    @raises(ValueError)
    def test_badvalue_choices(self):
        self.s['change_level'] = 'bad choice'

    @raises(ValueError)
    def test_emptyvalue_char(self):
        self.s['file_type'] = ''

    @raises(ValueError)
    def test_emptyvalue_char(self, key='file_type'):
        self.s[key] = ''  # Required field

    def test_emptyvalue_optional(self):
        self.s['additional_considerations'] = ''  # Optional field

    def test_resetfield(self, key='deliverer'):
        new_value = 'Wombat'
        self.s[key] = new_value
        assert_equals(self.s[key], new_value)
        del self.s[key]
        assert_not_equals(self.s[key], new_value)  # Also assumes KeyError is not thrown!

    def test_addfiles(self):
        for filename in self.tempfiles:
            self.s.add_file(filename)

    @raises(FileNotFoundError)
    def test_addbadfile(self):
        self.s.add_file(os.path.join(self.tmpdir, 'missing_file.fits'))

    def test_rmfile(self):
        for filename in self.tempfiles:
            self.s.add_file(filename)
        self.s.remove_file(list(self.tempfiles)[0])
        assert_equals( len(self.s.files), len(self.tempfiles)-1 )

    @raises(KeyError)
    def test_rmbadfile(self):
        for filename in self.tempfiles:
            self.s.add_file(filename)
        self.s.remove_file('bad_filename.fits')

    def test_yaml(self):
        assert_true(self.s.yaml)

    def test_help(self):
        self.s.help()  # Prints stuff

    @raises(ValueError)
    def test_validate_emptykey(self, key='file_type'):
        del self.s[key]  # Resets to empty str
        self.s.validate()

    @raises(NoFilesSelected)
    def test_validate_emptyfiles(self):
        for filename in self.s.files:
            self.s.remove_file(filename)

        # Do something here to pass field validation checks:
        self.s['file_type']           = 'value'
        self.s['correctness_testing'] = 'value'
        self.s['deliverer']           = 'value'
        self.s['description']         = 'value'
        self.s['calpipe_version']     = 'value'
        self.s['modes_affected']      = 'value'
        self.s['instrument']          = 'stis'  # Only works for HST

        self.s.validate()

    def test_validate(self):
        self.s.add_file(list(self.tempfiles)[0])

        # Do something here to pass field validation checks:
        self.s['file_type']           = 'value'
        self.s['correctness_testing'] = 'value'
        self.s['deliverer']           = 'value'
        self.s['description']         = 'value'
        self.s['calpipe_version']     = 'value'
        self.s['modes_affected']      = 'value'
        self.s['instrument']          = 'stis'  # Only works for HST

        self.s.validate()



'''
Created on Feb 15, 2017

@author: jmiller
'''
import re
import os.path

# ================================================================================================================

from .abstract import AbstractFile

# ================================================================================================================

def is_geis(name):
    """Return True IFF `name` identifies a GEIS header file."""
    name = os.path.basename(name)
    return bool(re.match(r"r[0-9](h|d)", name[-3:]))

def is_geis_data(name):
    """Return True IFF `name` identifies a GEIS data file."""
    name = os.path.basename(name)
    return bool(re.match(r"r[0-9]d", name[-3:]))

def is_geis_header(name):
    """Return True IFF `name` identifies a GEIS data file."""
    name = os.path.basename(name)
    return bool(re.match(r"r[0-9]h", name[-3:]))

def get_conjugate(reference):
    """Return any file associated with `reference`,  nominally GEIS data from header,
    e.g.  'something.r3h' --> 'something.r3d'
    If there is no file associated with reference,  return None.
    """
    if is_geis_data(reference):
        return reference[:-1] + "h"
    elif is_geis_header(reference):
        return reference[:-1] + "d"
    return None

# ================================================================================================================

class GeisFile(AbstractFile):

    format = "GEIS"

    def get_raw_header(self, needed_keys=(), **keys):
        """Return the header dictionary containing `needed_keys` from GEIS file at `name`."""

        filepath = self.filepath
        if isinstance(filepath, str):
            if filepath.endswith("d"):
                filepath = filepath[:-1] + "h"
            with open(filepath) as pfile:
                lines = pfile.readlines()
        else:  # assume file-like object
            lines = filepath

        header = {}
        history = []

        for line in lines:
            # Drop comment
            if len(line) >= 32 and line[31] == "/":
                line = line[:31]

            if line.startswith("HISTORY"):
                history.append(str(line[len("HISTORY"):].strip()))
                continue

            words = [x.strip() for x in line.split("=")]

            if len(words) < 2:
                continue

            key = words[0]

            # Skip over unneeded keys
            if needed_keys and key not in needed_keys:
                continue

            # Recombine value / comment portion
            value = "=".join(words[1:])

            # Remove quotes from strings
            value = value.strip()
            if value and value[0] == "'" and value[-1] == "'":
                value = value[1:-1].strip()

            # Assign value,  supporting list of values for HISTORY
            header[str(key)] = str(value)

        if not needed_keys or "HISTORY" in needed_keys:
            header["HISTORY"] = "\n".join(history)

        return header

_GEIS_TEST_DATA = u"""
SIMPLE  =                    F /

BITPIX  =                   16 /
DATATYPE= 'INTEGER*2'          /
NAXIS   =                    1 /
NAXIS1  =                  800 /
GROUPS  =                    T /
GCOUNT  =                    4 /
PCOUNT  =                   38 /
PSIZE   =                 1760 /
PTYPE1  = 'CRVAL1  '           /right ascension of reference pixel
PDTYPE1 = 'REAL*8  '           /
PSIZE1  =                   64 /
PTYPE2  = 'CRVAL2  '           /declination of reference pixel
PDTYPE2 = 'REAL*8  '           /
PSIZE2  =                   64 /

                               / GROUP PARAMETERS: OSS


                               / GROUP PARAMETERS: PODPS

INSTRUME= 'WFPC2   '           / instrument in use
ROOTNAME= 'F8213081U'          / rootname of the observation set
FILETYPE= 'MSK     '           / shp, ext, edq, sdq, sci

                               / SCIENCE INSTRUMENT CONFIGURATION

FILTNAM1= '        '           / first filter name
FILTNAM2= '        '           / second filter name
FILTER1 =                    0 / first filter number (0-48)
FILTER2 =                    0 / second filter number (0-48)

UCH1CJTM=                   0. / TEC cold junction #1 temperature (Celcius)
UCH2CJTM=                   0. / TEC cold junction #2 temperature (Celcius)
UCH3CJTM=                   0. / TEC cold junction #3 temperature (Celcius)
UCH4CJTM=                   0. / TEC cold junction #4 temperature (Celcius)
UBAY3TMP=                   0. / Bay 3 A1 temperature (Celcius)
KSPOTS  = 'OFF     '           / Status of Kelsall spot lamps: ON, OFF
SHUTTER = '        '           / Shutter in place during preflash or IFLAT (A,B)
ATODGAIN=                   0. /

                               / RSDP CONTROL KEYWORDS

PEDIGREE= 'INFLIGHT 01/01/1994 - 15/05/1995'
DESCRIP = 'STATIC MASK - INCLUDES CHARGE TRANSFER TRAPS'
HISTORY This file was edited by Michael S. Wiggs, August 1995
HISTORY
HISTORY e2112084u.r0h was edited to include values of 256
END
"""

# ------ prevent string join

"""
>>> is_geis("foo.r0h")
True
>>> is_geis("bar.fits")
False

>>> from io import StringIO
>>> header = get_geis_header(StringIO(_GEIS_TEST_DATA))

>> import pprint
>> pprint.pprint(header)
    {'ATODGAIN': '0.',
     'BITPIX': '16',
     'DATATYPE': 'INTEGER*2',
     'DESCRIP': 'STATIC MASK - INCLUDES CHARGE TRANSFER TRAPS',
     'FILETYPE': 'MSK',
     'FILTER1': '0',
     'FILTER2': '0',
     'FILTNAM1': '',
     'FILTNAM2': '',
     'GCOUNT': '4',
     'GROUPS': 'T',
     'HISTORY': 'This file was edited by Michael S. Wiggs, August 1995\n'
                '\n'
                'e2112084u.r0h was edited to include values of 256',
     'INSTRUME': 'WFPC2',
     'KSPOTS': 'OFF',
     'NAXIS': '1',
     'NAXIS1': '800',
     'PCOUNT': '38',
     'PDTYPE1': 'REAL*8',
     'PDTYPE2': 'REAL*8',
     'PEDIGREE': 'INFLIGHT 01/01/1994 - 15/05/1995',
     'PSIZE': '1760',
     'PSIZE1': '64',
     'PSIZE2': '64',
     'PTYPE1': 'CRVAL1',
     'PTYPE2': 'CRVAL2',
     'ROOTNAME': 'F8213081U',
     'SHUTTER': '',
     'SIMPLE': 'F',
     'UBAY3TMP': '0.',
     'UCH1CJTM': '0.',
     'UCH2CJTM': '0.',
     'UCH3CJTM': '0.',
     'UCH4CJTM': '0.'}
"""

def test():
    """Run doctest on data_file module."""
    import doctest
    from . import geis
    return doctest.testmod(geis)

if __name__ == "__main__":
    print(test())



"""
usage: /Users/jmiller/work/workspace_crds/CRDS/crds/misc/uniqname.py
       [-h] [--files FILES [FILES ...]] [--dry-run] [-a] [-f] [-e] [-s] [-r]
       [-o OUTPUT_PATH] [-b] [--fits-errors] [-v] [--verbosity VERBOSITY]
       [--dump-cmdline] [-R] [-I] [-V] [-J] [-H] [--stats] [--profile PROFILE]
       [--log-time] [--pdb] [--debug-traps]

This script is used to rename references with unique official CRDS names.

optional arguments:
  -h, --help            show this help message and exit
  --files FILES [FILES ...]
                        Files to rename.
  --dry-run             Print how a file would be renamed without modifying it.
  -a, --add-checksum    Add FITS checksum.  Without, checksums *removed* if header modified.
  -f, --add-keywords    When renaming, add FILENAME, ROOTNAME, HISTORY keywords for the generated name.
  -e, --verify-file     Verify FITS compliance when changing files.
  -s, --standard        Same as --add-keywords --verify-file,  does not add checksums (add -a).
  -r, --remove-original
                        After renaming,  remove the orginal file.
  -o OUTPUT_PATH, --output-path OUTPUT_PATH
                        Output renamed files to this directory path.
  -b, --brief           Produce less output.
  --fits-errors         When set, treat FITS compliance and checksum errors as fatal exceptions.
  -v, --verbose         Set log verbosity to True,  nominal debug level.
  --verbosity VERBOSITY
                        Set log verbosity to a specific level: 0..100.
  --dump-cmdline        Dump the command line parameters used to start the script to the log.
  -R, --readonly-cache  Don't modify the CRDS cache.  Not compatible with options which implicitly modify the cache.
  -I, --ignore-cache    Download required files even if they're already in the cache.
  -V, --version         Print the software version and exit.
  -J, --jwst            Force observatory to JWST for determining header conventions.
  -H, --hst             Force observatory to HST for determining header conventions.
  --stats               Track and print timing statistics.
  --profile PROFILE     Output profile stats to the specified file.
  --log-time            Add date/time to log messages.
  --pdb                 Run under pdb.
  --debug-traps         Bypass exception error message traps and re-raise exception.

This program is based loosely on the CDBS program uniqname modified to support
enhanced CDBS-style names with modified timestamps valid after 2016-01-01.

The CRDS uniqame is nominally run as follows:

    % crds uniqname --files s7g1700gl_dead.fits --brief --standard
    CRDS - INFO - Rewriting 's7g1700gl_dead.fits' --> 'zc52141pl_dead.fits'

If -s or --standard is added then routinely used switches are added as a
predefined bundle.   Initially these are --add-keywords and --verify-file.

If --add-checksum is specified,  CRDS uniqname will add FITS checksums to the file.
If --add-checksum is not specified,  CRDS uniqname WILL REMOVE any existing checksum.

If --verify-file is specified,  CRDS uniqname will check the FITS checksum and validate
the FITS format of renamed files.

If  --add-keywords is specified CRDS uniqname will add/modify the FILENAME, ROOTNAME,
and HISTORY to document the renaming.

If --remove-original is specified then the original file is deleted after the renamed
file has been created and modified as specified (checksums, keywords, etc.)

Renamed files can be output to a different directory using --output-path.

--dry-run can be used to demo renaming by printing what the new name would be.

"""
import os
import doctest

from crds.core import log
from crds import tests
from crds.tests import test_config

from crds.misc import uniqname
from crds.misc.uniqname import UniqnameScript

def dt_synphot_uniqname():
    """
    Compute diffs for two .pmap's:

    >>> old_state = test_config.setup()
    >>> name = UniqnameScript("crds.misc.uniqname --dry-run --files data/16n1832tm_tmc.fits")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Would rename 'data/16n1832tm_tmc.fits' --> 'data/...m_tmc.fits'
    >>> test_config.cleanup(old_state)
    """

def dt_cdbs_uniqname():
    """
    Compute diffs for two .pmap's:

    >>> old_state = test_config.setup()
    >>> name = UniqnameScript("crds.misc.uniqname --standard --files data/s7g1700gl_dead.fits")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Rewriting 'data/s7g1700gl_dead.fits' --> 'data/..._dead.fits'
    >>> name1 = uniqname.uniqname(name)
    CRDS - INFO -  Rewriting '..._dead.fits' --> '..._dead.fits'
    >>> os.remove(name1)
    >>> test_config.cleanup(old_state)
    """

def dt_has_checksum():
    """
    Compute diffs for two .pmap's:

    >>> old_state = test_config.setup()
    >>> uniqname.has_checksum("data/16n1832tm_tmc.fits")  # doctest: +ELLIPSIS
    False
    >>> uniqname.has_checksum("data/s7g1700gl_dead_good_xsum.fits") # doctest: +ELLIPSIS
    True
    >>> test_config.cleanup(old_state)
    """

def test():
    """Run module tests,  for now just doctests only.

    test_config.setup() and cleanup() are done inline above because bracketing
    the tests here does not get picked up by nose test discovery.  Combining
    tests into one giant docstring works but is hard to analyze and debug when
    things go wrong.
    """
    from crds.tests import test_uniqname, tstmod
    return tstmod(test_uniqname)

if __name__ == "__main__":
    print(test())



import os
import subprocess
import unittest
import pathlib


from crds.core import heavy_client
from crds.core.exceptions import CrdsLookupError

from crds.tests import test_config
from nose.tools import raises


class TestRoman(unittest.TestCase):
    """ TestRoman is a collection of crds unit tests for ROMAN.
    """
    def setUp(self):
        self.old_state = test_config.setup(
            url="https://roman-crds-serverless.stsci.edu",
            observatory="roman",
            cache=test_config.CRDS_TESTING_CACHE
        )

    def tearDown(self):
        test_config.cleanup(self.old_state)

    def test_getreferences_with_valid_header(self):
        """ test_getreferences_with_valid_header: test satisfies Roman 303.1 and 628.1
        """
        result = heavy_client.getreferences(
            {
                "ROMAN.META.INSTRUMENT.NAME": "WFI",
                "ROMAN.META.INSTRUMENT.DETECTOR": "WFI01",
                "ROMAN.META.EXPOSURE.TYPE": "WFI_IMAGE",
                "ROMAN.META.OBSERVATION.DATE": "2020-02-01",
                "ROMAN.META.OBSERVATION.TIME": "00:00:00",
            },
            observatory="roman",
            context="roman_0005.pmap",
            reftypes=["dark"]
        )

        assert pathlib.Path(result["dark"]).name == "roman_wfi_dark_0001.asdf"

        result = heavy_client.getreferences(
            {
                "ROMAN.META.INSTRUMENT.NAME": "WFI",
                "ROMAN.META.INSTRUMENT.DETECTOR": "WFI01",
                "ROMAN.META.EXPOSURE.TYPE": "WFI_GRISM",
                "ROMAN.META.INSTRUMENT.OPTICAL_ELEMENT": "GRISM",
                "ROMAN.META.OBSERVATION.DATE": "2020-02-01",
                "ROMAN.META.OBSERVATION.TIME": "00:00:00",
            },
            observatory="roman",
            context="roman_0005.pmap",
            reftypes=["flat"]
        )

        assert pathlib.Path(result["flat"]).name == "roman_wfi_flat_0004.asdf"


    @raises(CrdsLookupError)
    def test_getreferences_with_invalid_header(self):
        """ test_getreferences_with_invalid_header: test satisfies Roman 303.1
        """
        heavy_client.getreferences(
            {
                "ROMAN.META.INSTRUMENT.NAME": "WFI",
                "ROMAN.META.INSTRUMENT.DETECTOR": "WFI02",
                "ROMAN.META.EXPOSURE.TYPE": "WFI_IMAGE",
                "ROMAN.META.OBSERVATION.DATE": "2020-02-01",
                "ROMAN.META.OBSERVATION.TIME": "00:00:00",
            },
            observatory="roman",
            context="roman_0005.pmap",
            reftypes=["dark"]
        )

    def test_list_references(self):
        """ test_list_references: test satisfies Roman 303.2 and 628.2
        """
        env = os.environ.copy()

        expected_result = {
            "roman_wfi_flat_0004.asdf",
            "roman_wfi_dark_0001.asdf",
        }

        list_command = [
            'crds',
            'list',
            '--references',
            '--contexts',
            'roman_wfi_0004.imap'
        ]

        p = subprocess.Popen(
            list_command,
            env=env,
            stdout=subprocess.PIPE

        )
        results, err = p.communicate()
        results = results.decode('ascii').split("\n")

        assert {item for item in results if item} == expected_result



import os

import crds
from crds.core import log, utils, config
from crds import client
from crds.bestrefs import BestrefsScript

from crds.tests import test_config

# Contrived headers which will select bad files.

HST_HEADER = {
    'INSTRUME' : 'ACS',
    'REFTYPE' : 'PFLTFILE',
    'DETECTOR': 'SBC',
    'CCDAMP': 'N/A',
    'FILTER1' : 'PR110L',
    'FILTER2' : 'N/A',
    'OBSTYPE': 'SPECTROSCOPIC',
    'FW1OFFST' : 'N/A',
    'FW2OFFST': 'N/A',
    'FWOFFST': 'N/A',
    'DATE-OBS': '1991-01-01',
    'TIME-OBS': '00:00:00'
    }

JWST_HEADER = {
    "meta.instrument.name": "MIRI",
    "meta.observation.date": "2012-07-25T00:00:00",
    "meta.instrument.detector" : "MIRIMAGE",
    "meta.instrument.filter" : "F1000W",
    "meta.subarray.name" : "FULL",
    }

def dt_bad_references_error_cache_config():
    """
    CRDS can designate files as scientifically invalid which is reflected in the catalog
    on a the CRDS server and also recorded in the configuration info and as a bad files list
    which are written down in the "config" directory.

    A key aspect of bad files management is the location and contents of the cache config
    directory.  The current HST cache in trunk/crds/cache has a config area and 4 bad files.

    The default handling when a bad reference file is assigned is to raise an exception:

    >>> old_state = test_config.setup(clear_existing=False)
    >>> config.ALLOW_BAD_RULES.reset()
    >>> config.ALLOW_BAD_REFERENCES.reset()

    >>> crds.getreferences(HST_HEADER, observatory='hst', context='hst_0282.pmap', reftypes=['pfltfile'])
    Traceback (most recent call last):
    ...
    CrdsBadReferenceError: Recommended reference 'l2d0959cj_pfl.fits' of type 'pfltfile' is designated scientifically invalid.
    <BLANKLINE>

    >>> test_config.cleanup(old_state)
    """

def dt_bad_references_warning_cache_config():
    """
    A secondary behaviour is to permit use of bad references with a warning:

    >>> old_state = test_config.setup(clear_existing=False)
    >>> config.ALLOW_BAD_RULES.set("1")
    False
    >>> config.ALLOW_BAD_REFERENCES.set("1")
    False

    >>> crds.getreferences(HST_HEADER, observatory='hst', context='hst_0282.pmap', reftypes=['pfltfile'])    # doctest: +ELLIPSIS
    CRDS - WARNING - Recommended reference 'l2d0959cj_pfl.fits' of type 'pfltfile' is designated scientifically invalid.
    <BLANKLINE>
    {'pfltfile': '.../l2d0959cj_pfl.fits'}

    >>> test_config.cleanup(old_state)
    """

def dt_bad_references_fast_mode():
    """
    When run in 'fast' mode as is done for the calls from crds.bestrefs,  no exception or warning is possible:

    >>> old_state = test_config.setup(clear_existing=False)
    >>> config.ALLOW_BAD_RULES.reset()
    >>> config.ALLOW_BAD_REFERENCES.reset()

    >>> crds.getreferences(HST_HEADER, observatory='hst', context='hst_0282.pmap', reftypes=['pfltfile'], fast=True) # doctest: +ELLIPSIS
    {'pfltfile': '.../references/hst/l2d0959cj_pfl.fits'}

    >>> test_config.cleanup(old_state)
    """


def dt_bad_references_bestrefs_script_error():
    """
    The crds.bestrefs program handles bad files differently because it frequently operates on
    multiple contexts at the same time,  and recommending bad files under the old context is OK.

    >>> old_state = test_config.setup(clear_existing=False)
    >>> config.ALLOW_BAD_RULES.reset()
    >>> config.ALLOW_BAD_REFERENCES.reset()

    By default,  in crds.bestrefs use of a bad reference is an error:

    >>> BestrefsScript("crds.bestrefs --new-context hst_0282.pmap --files data/j8btxxx_raw_bad.fits")()
    CRDS - INFO - No comparison context or source comparison requested.
    CRDS - INFO - No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO - ===> Processing data/j8btxxx_raw_bad.fits
    CRDS - ERROR - instrument='ACS' type='PFLTFILE' data='data/j8btxxx_raw_bad.fits' ::  File 'L2D0959CJ_PFL.FITS' is bad. Use is not recommended,  results may not be scientifically valid.
    CRDS - INFO - 1 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 3 infos
    1

    >>> test_config.cleanup(old_state)
    """

def dt_bad_references_bestrefs_script_warning():
    """
    >>> old_state = test_config.setup(clear_existing=False)
    >>> config.ALLOW_BAD_RULES.set("1")
    False
    >>> config.ALLOW_BAD_REFERENCES.set("1")
    False

    >>> BestrefsScript("crds.bestrefs --new-context hst_0282.pmap --files data/j8btxxx_raw_bad.fits --allow-bad-references")() # doctest: +ELLIPSIS
    CRDS - INFO - No comparison context or source comparison requested.
    CRDS - INFO - No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO - ===> Processing data/j8btxxx_raw_bad.fits
    CRDS - WARNING - For data/j8btxxx_raw_bad.fits ACS pfltfile File 'L2D0959CJ_PFL.FITS' is bad. Use is not recommended,  results may not be scientifically valid.
    CRDS - INFO - 0 errors
    CRDS - INFO - 1 warnings
    CRDS - INFO - 3 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_bad_references_bestrefs_script_deprecated():
    """
    >>> old_state = test_config.setup(clear_existing=False)
    >>> config.ALLOW_BAD_RULES.reset()
    >>> config.ALLOW_BAD_REFERENCES.reset()

    >>> BestrefsScript("crds.bestrefs --new-context hst_0282.pmap --files data/j8btxxx_raw_bad.fits")() # doctest: +ELLIPSIS
    CRDS - INFO - No comparison context or source comparison requested.
    CRDS - INFO - No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.
    CRDS - INFO - ===> Processing data/j8btxxx_raw_bad.fits
    CRDS - ERROR - instrument='ACS' type='PFLTFILE' data='data/j8btxxx_raw_bad.fits' ::  File 'L2D0959CJ_PFL.FITS' is bad. Use is not recommended,  results may not be scientifically valid.
    CRDS - INFO - 1 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 3 infos
    1

    >>> test_config.cleanup(old_state)
    """

def dt_bad_rules_jwst_getreferences_error():
    """
    There is also a check for use of bad rules. JWST has a few,  including jwst_0017.pmap by "inheritance"
    since it includes some bad rules.

    Do some setup to switch to a JWST serverless mode.

    >>> old_state = test_config.setup(cache=test_config.CRDS_SHARED_GROUP_CACHE, url="https://jwst-serverless-mode.stsci.edu")
    >>> config.ALLOW_BAD_RULES.reset()
    >>> config.ALLOW_BAD_REFERENCES.reset()

    >>> crds.getreferences(JWST_HEADER, observatory='jwst', context='jwst_0017.pmap', reftypes=["flat"])
    Traceback (most recent call last):
    ...
    CrdsBadRulesError: Final context 'jwst_0017.pmap' is marked as scientifically invalid based on: ['jwst_miri_flat_0003.rmap']
    <BLANKLINE>

    >>> test_config.cleanup(old_state)
    """

def dt_bad_rules_jwst_getreferences_warning():
    """
    Similarly,  the use of bad rules can be permitted:

    >>> old_state = test_config.setup(cache=test_config.CRDS_SHARED_GROUP_CACHE, url="https://jwst-serverless-mode.stsci.edu")
    >>> config.ALLOW_BAD_RULES.set("1")
    False

    >>> refs = crds.getreferences(JWST_HEADER, observatory='jwst', context='jwst_0017.pmap', reftypes=["flat"])   # doctest: +ELLIPSIS
    CRDS - WARNING - Final context 'jwst_0017.pmap' is marked as scientifically invalid based on: ['jwst_miri_flat_0003.rmap']
    <BLANKLINE>

    >>> list(refs.keys()) == ['flat']
    True

    >>> os.path.basename(refs['flat'])
    'jwst_miri_flat_0006.fits'

    >>> test_config.cleanup(old_state)
    """

def dt_bad_rules_jwst_bestrefs_script_error():
    """
    Here try bad rules for a JWST dataset:

    >>> old_state = test_config.setup(cache=test_config.CRDS_SHARED_GROUP_CACHE, url="https://jwst-serverless-mode.stsci.edu")
    >>> config.ALLOW_BAD_RULES.reset()

    >>> crds.getrecommendations(JWST_HEADER, reftypes=["gain"], context="jwst_0017.pmap")    # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    CrdsBadRulesError: Final context 'jwst_0017.pmap' is marked as scientifically invalid based on: ['jwst_miri_flat_0003.rmap']
    <BLANKLINE>

    >>> test_config.cleanup(old_state)
    """

def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_bad_files, tstmod
    return tstmod(test_bad_files)

if __name__ == "__main__":
    print(main())



"""This module provides a command line interface for doing CRDS file submissions."""
import sys
import os
import os.path
import time

from crds.core import log, config, utils, cmdline, exceptions
from crds.core.log import srepr
from crds.core.exceptions import CrdsError
from . import web, background, monitor
from crds.client import api
from crds.certify import certify

# ===================================================================

class ReferenceSubmissionScript(cmdline.Script):
    """Command line script file submission script."""

    description = """
This script initiates a CRDS file submission from the command line rather than the
web server.   Once submitted the submission will be processed on the CRDS server.
"""

    epilog = """
Typical usage of this script would be something like the following.  Users of
this command line interface must be members of the CRDS operators group
"crdsoper".
"""

    # Override CRDS default behavior of looking in cache for path-less files.
    locate_file = cmdline.Script.locate_file_outside_cache

    def __init__(self, *args, **keys):

        """Initializes a reference file submission script."""

        super(ReferenceSubmissionScript, self).__init__(*args, **keys)

        self.username = None
        self.connection = None
        self.submission = None
        self.pmap_mode = None
        self.pmap_name = None
        self.instruments_filekinds = None
        self.instrument = None
        self.base_url = None
        self.jpoll_key = None

        self._ready_url = None
        self._error_count = None
        self._warning_count = None

    def add_args(self):
        """Add additional command-line parameters for file submissions not found in baseclass Script."""
        super(ReferenceSubmissionScript, self).add_args()
        self.add_argument("--files", nargs="*", help="Files to submit.  A file preceded with @ is treated as containing the list of files.")
        self.add_argument("--derive-from-context", type=cmdline.context_spec, default=None,
                          help="Set of CRDS rules these files will be added to.  Defaults to edit context.")
        self.add_argument("--change-level", type=str, choices=["SEVERE","MODERATE","TRIVIAL"], default="SEVERE",
                          help="The degree to which the new files are expected to impact science results.")
        self.add_argument("--creator", type=str,
                          help="Author of this set of references,  most likely not file submitter.  Can be comma separated list of people in quotes.")
        self.add_argument("--description", type=str, default=None,
                          help="Brief description of the purpose of this delivery, mention instrument and type(s).")
        self.add_argument("--dont-auto-rename", action="store_true",
                          help="Unless specified, CRDS will automatically rename incoming files.")
        self.add_argument("--dont-compare-old-reference", action="store_true",
                          help="Unless specified, CRDS will check the current reference against any reference it replaces, as appropriate and possible.")
        self.add_argument("--username", type=str, default=None, help="CRDS username of file submitter.")
        self.add_argument("--monitor-processing", action="store_true",
                          help="Monitor CRDS processing for on-going status and final confirmation link.")
        self.add_argument("--wait-for-completion", action="store_true",
                          help="Wait until the server reports that the submission is done before exiting.  Otherwise use e-mail.")
        self.add_argument("--submission-kind", type=str, choices=["batch","certify","none"], default="batch", # mapping, reference
                          help="Which form of submission to perform.  Defaults to batch.")
        self.add_argument("--wipe-existing-files", action="store_true",
                          help="Before performing action,  remove all files from the appropriate CRDS ingest directory.")
        self.add_argument("--keep-existing-files", action="store_true",
                          help="Don't recopy files already in the server ingest directory that have the correct length.")
        self.add_argument("--certify-files", action="store_true",
                          help="Run CRDS certify and fail if errors are found.")

        self.add_argument("--logout", action="store_true",
                          help="Log out of the server,  dropping any lock.")

    # -------------------------------------------------------------------------------------------------

    def finish_parameters(self):
        """Finish up parameter setup which requires parsed command line arguments."""
        self.username = self.args.username or config.get_username()
        password = config.get_password()
        self.base_url = config.get_server_url(self.observatory)
        self.instruments_filekinds = utils.get_instruments_filekinds(self.observatory, self.files) if self.args.files else {}
        self.instrument = list(self.instruments_filekinds.keys())[0] if len(self.instruments_filekinds) == 1 else "none"
        self.connection = web.CrdsDjangoConnection(
            locked_instrument=self.instrument, username=self.username, password=password, base_url=self.base_url)
        if self.args.derive_from_context is None:
            self.args.derive_from_context = self.observatory + "-edit"
        if self.args.derive_from_context.endswith(("-edit", "-operational")):
            # this actually floats for concurrent deliveries
            self.pmap_mode = "pmap_" + self.args.derive_from_context.split("-")[-1]
        else:
            self.pmap_mode = "pmap_text"
        self.pmap_name = self.resolve_context(self.args.derive_from_context)
        assert config.is_context(self.pmap_name), "Invalid pmap_name " + repr(self.pmap_name)
        assert not (self.args.keep_existing_files and self.args.wipe_existing_files), \
            "--keep-existing-files and --wipe-existing-files are mutually exclusive."

    # -------------------------------------------------------------------------------------------------

    def ingest_files(self):
        """Upload self.files to the user's ingest directory on the CRDS server."""
        stats = self._start_stats()
        total_size = utils.total_size(self.files)

        ingest_info = self.get_ingested_files()

        self.scan_for_nonsubmitted_ingests(ingest_info)

        remaining_files = self.keep_existing_files(ingest_info, self.files) \
            if self.args.keep_existing_files else self.files

        for i, filename in enumerate(remaining_files):
            file_size = utils.file_size(filename)
            log.info("Upload started", repr(filename), "[", i+1, "/", len(self.files), " files ]",
                     "[", utils.human_format_number(file_size),
                     "/", utils.human_format_number(total_size), " bytes ]")
            self.connection.upload_file(filename)
            stats.increment("bytes", file_size)
            stats.increment("files", 1)
            stats.log_status("files", "Upload complete", len(self.files))
            stats.log_status("bytes", "Upload complete", total_size)

        log.divider(func=log.verbose)
        stats.report()
        log.divider(char="=")

    def scan_for_nonsubmitted_ingests(self, ingest_info):
        """Check for junk in the submitter's ingest directory,  left over files not
        in the current submission and fail if found.
        """
        submitted_basenames = [ os.path.basename(filepath) for filepath in self.files ]
        msg = None
        for ingested in ingest_info.keys():
            if ingested not in submitted_basenames:
                msg = log.format("Non-submitted file", log.srepr(ingested),
                                 "is already in the CRDS server's ingest directory.  Delete it (--wipe-existing-files or web page Upload Files panel) or submit it.")
                log.error(msg)
        if msg is not None:
            raise exceptions.CrdsExtraneousFileError(
                "Unexpected files already delivered to CRDS server. See ERROR messages.")

    def keep_existing_files(self, ingest_info, files):
        """Keep files which have already been copied and have the correct server side
        length.  This can save *hours* of copy time for repeat submissions.
        """
        for filename in files[:]:
            local_size = utils.file_size(filename)
            basename = os.path.basename(filename)
            try:
                existing_size = int(ingest_info[basename]["size"])
            except Exception:
                log.info("File", repr(filename),
                         "does not exist in ingest directory and will be copied to CRDS server.")
                continue
            if local_size == existing_size:
                log.info("File", repr(filename),
                         "has already been copied and has correct length on CRDS server",
                         utils.human_format_number(existing_size))
                files.remove(filename)
            else:
                log.info("File", repr(filename),
                         "exists but has incorrect size and must be recopied.  Deleting old ingest.")
                self.connection.get(ingest_info[basename]["deleteUrl"])
        return files

    def get_ingested_files(self):
        """Return the server-side JSON info on the files already in the submitter's ingest directory."""
        log.verbose("Querying for existing files.")
        result = self.connection.get('/upload/list/').json()
        log.verbose("JSON info on existing ingested files:\n", log.PP(result))
        if "files" in result and isinstance(result["files"], list):
            return { info["name"] : info for info in result["files"] }
        return { info["name"] : info for info in result }

    def wipe_files(self):
        """Delete all files from the user's ingest directory on the CRDS server."""
        log.divider(name="wipe files", char="=")
        ingest_info = self.get_ingested_files()
        for basename in ingest_info:
            log.info("Wiping file", repr(basename))
            self.connection.get(ingest_info[basename]["deleteUrl"])

    def _start_stats(self):
        """Helper method to initialize stats keeping for ingest."""
        total_bytes = utils.total_size(self.files)
        stats = utils.TimingStats(output=log.verbose)
        stats.start()
        log.divider(name="ingest files", char="=")
        log.info("Uploading", len(self.files), "file(s) totalling", utils.human_format_number(total_bytes), "bytes")
        log.divider(func=log.verbose)
        return stats

    # -------------------------------------------------------------------------------------------------

    def jpoll_open_channel(self):
        """Mimic opening a JPOLL status channel as do pages with real-time status."""
        log.info("Preparing server logging.")
        response = self.connection.get("/jpoll/open_channel")
        return response.json()

    # -------------------------------------------------------------------------------------------------

    def certify_files(self):
        """Run the CRDS server Certify Files page on `filepaths`."""
        self.ingest_files()
        t = self.connection.repost_start(
            "/certify/", pmap_name=self.pmap_name, pmap_mode=self.pmap_mode,
            compare_old_reference=not self.args.dont_compare_old_reference)
        time.sleep(10)
        return t

    # -------------------------------------------------------------------------------------------------

    def batch_submit_references(self):
        """Do a web re-post to the batch submit references web page."""
        return self._submission("/batch_submit_references/")

    def submit_references(self):
        """Do a web re-post to the submit references web page."""
        return self._submission("/submit/reference/")

    def submit_mappings(self):
        """Do a web re-post to the submit mappings web page."""
        return self._submission("/submit/mapping/")

    def _submission(self, relative_url):
        """Do a generic submission re-post to the specified relative_url."""
        assert self.args.description is not None, "You must supply a --description for this function."
        self.ingest_files()
        log.info("Posting web request for", srepr(relative_url))
        submission_args = self.get_submission_args()
        completion_args = self.connection.repost_start(relative_url, **submission_args)
        # give POST time to complete send, not response
        time.sleep(10)
        return completion_args

    def get_submission_args(self):
        """Return a dictionary mapping form variables to value strings for the basic command
        line submission parameters.
        """
        result = dict(
            pmap_mode = self.pmap_mode,
            pmap_name = self.pmap_name,
            instrument = self.instrument,
            change_level = self.args.change_level,
            creator = self.args.creator,
            description = self.args.description,
            auto_rename = not self.args.dont_auto_rename,
            compare_old_reference = not self.args.dont_compare_old_reference,
        )

        if self.pmap_mode == "pmap_text":
            result["pmap_text"] = self.pmap_name

        return result

    def submission_complete(self, args):
        """Threaded completion function for any submission,  returns web response."""
        return self.connection.repost_complete(args)

    @background.background
    def monitor(self):
        """Run a background job to monitor the submission on the server and output log info."""
        extra_params = ""
        if "--log-time" in sys.argv:
            extra_params = "--log-time"
        submission_monitor = monitor.MonitorScript("crds.monitor --key {} --poll {} {}".format(
            self.jpoll_key, 3, extra_params), reset_log=False)
        submission_monitor()
        return submission_monitor

    def monitor_complete(self, monitor_future):
        """Wait for the monitor job to complete and return the result."""
        return background.background_complete(monitor_future)

    # -------------------------------------------------------------------------------------------------

    def login(self):
        """Log in to the CRDS server using server user credentials."""
        log.info("Logging in aquiring lock.")
        self.connection.fail_if_existing_lock()
        self.connection.login()

    def logout(self):
        """Log out of the CRDS server,  releasing any lock held by this user."""
        log.info("Logging out releasing lock.")
        self.connection.login()
        self.connection.logout()

    def certify_local_files(self):
        """Run CRDS certify on submitted files and fail/exit if errors are
        found.
        """
        if log.errors():
            raise CrdsError("Errors encountered before CRDS certify run.")
        certify.certify_files(
            self.files, context=self.pmap_name, dump_provenance=True,
            compare_old_reference=True, observatory=self.observatory,
            run_fitsverify=True, check_rmap=False, check_sha1sums=True)
        if log.errors():
            raise CrdsError("Certify errors found.  Aborting submission.")

    def main(self):
        """Main control flow of submission directory and request manifest creation."""

        log.divider("setting up", char="=")

        self.require_server_connection()

        self.finish_parameters()

        if self.args.certify_files:
            self.certify_local_files()

        if self.args.logout:
            return self.logout()

        self.login()

        if self.args.wipe_existing_files:
            self.wipe_files()

        self.jpoll_key = self.jpoll_open_channel()

        if self.args.submission_kind == "batch":
            submit_future = self.batch_submit_references()
        elif self.args.submission_kind == "certify":
            submit_future = self.certify_files()
        elif self.args.submission_kind == "references":
            submit_future = self.submit_references()
        elif self.args.submission_kind == "mappings":
            submit_future = self.submit_mappings()

        if self.args.monitor_processing:
            monitor_future = self.monitor()

        if self.args.wait_for_completion:
            self.submission_complete(submit_future)

        if self.args.monitor_processing:
            monitor = self.monitor_complete(monitor_future)
            if monitor.exit_status == 0:
                self._ready_url = monitor.result

        log.standard_status()

        self._error_count = log.errors()
        self._warning_count = log.warnings()

        return log.errors()

# ===================================================================

# ===================================================================



"""This tests, through the use of bestrefs, the functioning of table effects."""
import doctest

from crds import tests
from crds.tests import test_config
from crds.bestrefs import BestrefsScript

def dt_table_effects_default_always_reprocess():
    """
    Test: Default rule: always reprocess, based on STIS PCTAB.
    Test: STIS APERTAB: No reprocess

    >>> old_state = test_config.setup()

    >>> doctest.ELLIPSIS_MARKER = '...'
    >>> BestrefsScript(argv="bestrefs.py -z  --verbosity 25 --old-context hst_0003.pmap  --new-context hst_0268.pmap --datasets O8EX02EFQ")() # doctest: +ELLIPSIS
    CRDS - DEBUG -  Using explicit new context 'hst_0268.pmap' for computing updated best references.
    CRDS - DEBUG -  Using explicit old context 'hst_0003.pmap'
    CRDS - INFO -  Dumping dataset parameters from CRDS server at 'https://...' for ['O8EX02EFQ']
    CRDS - INFO -  Dumped 1 of 1 datasets from CRDS server at 'https://...'
    CRDS - INFO -  Computing bestrefs for datasets ['O8EX02EFQ']
    CRDS - DEBUG -  ===> Processing O8EX02010:O8EX02EFQ
    CRDS - DEBUG -  Deep Reference examination between .../n7p1032ao_apt.fits and .../y2r1559to_apt.fits initiated.
    CRDS - DEBUG -  Instantiating rules for reference type stis_apertab.
    CRDS - DEBUG -  Rule DeepLook_STISaperture: Reprocessing is not required.
    CRDS - DEBUG -  Rule DeepLook_STISaperture: Selection rules have executed and the selected rows are the same.
    CRDS - DEBUG -  Removing table update for STIS apertab O8EX02010:O8EX02EFQ no effective change from reference 'N7P1032AO_APT.FITS' --> 'Y2R1559TO_APT.FITS'
    CRDS - DEBUG -  Deep Reference examination between .../q5417413o_pct.fits and .../y2r16006o_pct.fits initiated.
    CRDS - DEBUG -  Instantiating rules for reference type stis_pctab.
    CRDS - DEBUG -  Rule DeepLook_Default: Reprocessing is required.
    CRDS - DEBUG -  Rule DeepLook_Default: Reference type cannot be examined, by definition.
    CRDS - DEBUG -  1 sources processed
    CRDS - DEBUG -  1 source updates
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  3 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_table_effects_reprocess_test():
    """
    Test: COS WCPTAB, reprocess yes

    >>> old_state = test_config.setup()

    >>> doctest.ELLIPSIS_MARKER = '...'
    >>> BestrefsScript(argv="bestrefs.py -z  --verbosity 25 --old-context hst_0018.pmap  --new-context hst_0024.pmap --datasets LB6M01030")() # doctest: +ELLIPSIS
    CRDS - DEBUG -  Using explicit new context 'hst_0024.pmap' for computing updated best references.
    CRDS - DEBUG -  Using explicit old context 'hst_0018.pmap'
    CRDS - INFO -  Dumping dataset parameters from CRDS server at 'https://...' for ['LB6M01030']
    CRDS - INFO -  Dumped 1 of 1 datasets from CRDS server at 'https://...'
    CRDS - INFO -  Computing bestrefs for datasets ['LB6M01030']
    CRDS - DEBUG -  ===> Processing LB6M01030:LB6M01AVQ
    CRDS - DEBUG -  Deep Reference examination between .../x2i1559gl_wcp.fits and .../xaf1429el_wcp.fits initiated.
    CRDS - DEBUG -  Instantiating rules for reference type cos_wcptab.
    CRDS - DEBUG -  Rule DeepLook_COSOpt_elem: Reprocessing is required.
    CRDS - DEBUG -  Rule DeepLook_COSOpt_elem: Selection rules have executed and the selected rows are different.
    CRDS - DEBUG -  1 sources processed
    CRDS - DEBUG -  1 source updates
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  3 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_table_effects_reprocess_no():
    """
    Test: COS WCPTAB, reprocess no

    >>> old_state = test_config.setup()

    >>> doctest.ELLIPSIS_MARKER = '...'
    >>> BestrefsScript(argv="bestrefs.py -z  --verbosity 25 --old-context hst_0018.pmap  --new-context hst_0024.pmap --datasets LBK617YRQ")() # doctest: +ELLIPSIS
    CRDS - DEBUG -  Using explicit new context 'hst_0024.pmap' for computing updated best references.
    CRDS - DEBUG -  Using explicit old context 'hst_0018.pmap'
    CRDS - INFO -  Dumping dataset parameters from CRDS server at 'https://...' for ['LBK617YRQ']
    CRDS - INFO -  Dumped 1 of 1 datasets from CRDS server at 'https://...'
    CRDS - INFO -  Computing bestrefs for datasets ['LBK617YRQ']
    CRDS - DEBUG -  ===> Processing LBK617010:LBK617YRQ
    CRDS - DEBUG -  Deep Reference examination between .../x2i1559gl_wcp.fits and .../xaf1429el_wcp.fits initiated.
    CRDS - DEBUG -  Instantiating rules for reference type cos_wcptab.
    CRDS - DEBUG -  Rule DeepLook_COSOpt_elem: Reprocessing is not required.
    CRDS - DEBUG -  Rule DeepLook_COSOpt_elem: Selection rules have executed and the selected rows are the same.
    CRDS - DEBUG -  Removing table update for COS wcptab LBK617010:LBK617YRQ no effective change from reference 'X2I1559GL_WCP.FITS' --> 'XAF1429EL_WCP.FITS'
    CRDS - DEBUG -  1 sources processed
    CRDS - DEBUG -  0 source updates
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  3 infos
    0

    >>> test_config.cleanup(old_state)
    """

def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_table_effects, tstmod
    return tstmod(test_table_effects)

if __name__ == "__main__":
    print(main())



import os
import tempfile

# ==================================================================================
import numpy as np

from nose.tools import assert_raises, assert_true

# ==================================================================================

from crds.core import utils, log, exceptions
from crds.core import substitutions
from crds import hst, jwst

from crds.tests import test_config

# ==================================================================================

def substitutions_validate_hst():
    """
    >>> old_state = test_config.setup(url="https://hst-serverless-mode.stsci.edu")
    >>> old_verbose = log.set_verbose()
    >>> substitutions.validate_substitutions("hst-operational")
    CRDS - DEBUG -  Using CACHED CRDS reference assignment rules last updated on '...'
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 3710 3826' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 3826 3936' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 3936 4051' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 4051 4167' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 4167 4296' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 4296 4421' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 4421 4554' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 4554 4686' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 4686 4821' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 4821 5271' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 5271 5751' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 5751 6271' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 6271 6851' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 6851 7471' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LRFWAVE' with value 'between 7471 8161' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LTV1' with value 'ANY' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'LTV2' with value 'ANY' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'NAXIS1' with value 'ANY' is unchecked.
    CRDS - DEBUG -  For 'acs' parameter 'NAXIS2' with value 'ANY' is unchecked.
    CRDS - DEBUG -  Instrument 'nicmos' has no substitutions.
    CRDS - DEBUG -  Instrument 'synphot' has no substitutions.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'BLANK' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'CLEAR' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F093W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F098M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F105W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F110W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F125W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F126N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F127M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F128N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F130N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F132N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F139M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F140W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F153M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F160W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F164N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F167N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F200LP' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F218W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F225W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F275W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F280N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F300X' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F336W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F343N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F350LP' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F373N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F390M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F390W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F395N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F410M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F438W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F467M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F469N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F475W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F475X' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F487N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F502N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F547M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F555W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F588N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F600LP' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F606W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F621M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F625W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F631N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F645N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F656N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F657N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F658N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F665N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F673N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F680N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F689M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F763M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F775W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F814W' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F845M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F850LP' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'F953N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ232N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ243N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ378N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ387N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ422M' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ436N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ437N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ492N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ508N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ575N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ619N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ634N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ672N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ674N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ727N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ750N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ889N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ906N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ924N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'FILTER' with value 'FQ937N' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'MIF1200' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'MIF1500' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'MIF600' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'MIF900' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'NONE' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'RAPID' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'SPARS10' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'SPARS100' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'SPARS200' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'SPARS25' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'SPARS350' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'SPARS50' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'STEP100' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'STEP200' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'STEP25' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'STEP400' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'STEP50' is unchecked.
    CRDS - DEBUG -  For 'wfc3' parameter 'SAMP_SEQ' with value 'UNKNOWN' is unchecked.
    CRDS - DEBUG -  Instrument 'wfpc2' has no substitutions.
    >>> _ = log.set_verbose(old_verbose)
    >>> test_config.cleanup(old_state)
    """

def substitutions_validate_jwst():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> old_verbose = log.set_verbose()
    >>> substitutions.validate_substitutions("jwst-operational")  # doctest: +ELLIPSIS
    CRDS - DEBUG -  Using CACHED CRDS reference assignment rules last updated on '...'
    CRDS - DEBUG -  Instrument 'fgs' has no substitutions.
    CRDS - DEBUG -  Instrument 'miri' has no substitutions.
    CRDS - DEBUG -  Instrument 'nircam' has no substitutions.
    CRDS - DEBUG -  Instrument 'niriss' has no substitutions.
    CRDS - DEBUG -  Instrument 'nirspec' has no substitutions.
    CRDS - DEBUG -  Instrument 'system' has no substitutions.
    >>> _ = log.set_verbose(old_verbose)
    >>> test_config.cleanup(old_state)
    """

def substitutions_validate_roman():
    """
    >>> old_state = test_config.setup(url="https://roman-crds-serverless.stsci.edu", cache=test_config.CRDS_TESTING_CACHE)
    >>> old_verbose = log.set_verbose()
    >>> substitutions.validate_substitutions("roman-operational")  # doctest: +ELLIPSIS
    CRDS - DEBUG -  Using CACHED CRDS reference assignment rules last updated on '...'
    CRDS - DEBUG -  Instrument 'wfi' has no substitutions.
    >>> _ = log.set_verbose(old_verbose)
    >>> test_config.cleanup(old_state)
    """
# ==================================================================================

def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_substitutions, tstmod
    return tstmod(test_substitutions)

if __name__ == "__main__":
    print(main())



import os
import shutil
from contextlib import contextmanager
import warnings
from multiprocessing import Pool

from astropy.io import fits

from crds.core import rmap, log
from crds import config
from . import utils


class SynphotIntegrationTestBase:
    """
    Base class for synphot integration tests.
    """
    def __init__(
        self,
        context,
        graph_file=None,
        throughput_lookup_file=None,
        thermal_lookup_file=None,
        delivered_throughput_files=None,
        delivered_thermal_files=None
    ):
        self.context = context

        if graph_file is None:
            self.graph_file = utils.get_cache_path(context, utils.GRAPH_REFTYPE)
        else:
            self.graph_file = graph_file

        if throughput_lookup_file is None:
            self.throughput_lookup_file = utils.get_cache_path(context, utils.THROUGHPUT_LOOKUP_REFTYPE)
        else:
            self.throughput_lookup_file = throughput_lookup_file

        if thermal_lookup_file is None:
            self.thermal_lookup_file = utils.get_cache_path(context, utils.THERMAL_LOOKUP_REFTYPE)
        else:
            self.thermal_lookup_file = thermal_lookup_file

        if delivered_throughput_files is None:
            self.delivered_throughput_files = []
        else:
            self.delivered_throughput_files = delivered_throughput_files

        if delivered_thermal_files is None:
            self.delivered_thermal_files = []
        else:
            self.delivered_thermal_files = delivered_thermal_files

        self.imap = rmap.asmapping(context).get_imap(utils.SYNPHOT_INSTRUMENT)

        self.throughput_compname_to_path = self._build_compname_to_path(
            utils.THROUGHPUT_REFTYPE,
            self.delivered_throughput_files
        )

        self.thermal_compname_to_path = self._build_compname_to_path(
            utils.THERMAL_REFTYPE,
            self.delivered_thermal_files
        )

    def _build_compname_to_path(self, reftype, delivered_files):
        """
        Build a map of component names to filesystem paths, using freshly
        delivered files by preference.
        """
        result = {}
        for file in delivered_files:
            with fits.open(file) as hdul:
                # A significant minority of existing files have uppercase COMPNAME,
                # enough that it's not worthwhile to redeliver them all.
                result[hdul[0].header["COMPNAME"].lower()] = file

        mapping = self.imap.get_rmap(reftype)
        # This is a tad brittle, in that it assumes that the COMPNAME parkey
        # will occupy the first position.
        for key, filename in mapping.selector.todict()["selections"]:
            compname = key[0].lower()
            if not compname in result:
                result[compname] = config.locate_file(filename, "hst")

        return result


@contextmanager
def _configured_stsynphot(synphot_root):
    # stsynphot will complain that its Vega spectrum file is missing
    with warnings.catch_warnings():
        try:
            import stsynphot
        except ImportError:
            raise ImportError("Missing stsynphot package.  Install the 'synphot' extras and try again.")

    original_rootdir = stsynphot.conf.rootdir
    try:
        stsynphot.conf.rootdir = synphot_root
        yield stsynphot
    finally:
        stsynphot.conf.rootdir = original_rootdir


@contextmanager
def _configured_pysynphot(synphot_root):
    original_pysyn_cdbs = os.environ.get("PYSYN_CDBS")
    try:
        os.environ["PYSYN_CDBS"] = synphot_root
        try:
            import pysynphot
        except ImportError:
            raise ImportError("Missing pysynphot package.  Install the 'synphot' extras and try again.")

        yield pysynphot
    finally:
        if original_pysyn_cdbs is None:
            os.environ.pop("PYSYN_CDBS")
        else:
            os.environ["PYSYN_CDBS"] = original_pysyn_cdbs


def _test_synphot_mode(synphot_root, obsmode):
    result = True
    errors = []
    warns = []
    with _configured_stsynphot(synphot_root) as sts:
        try:
            with warnings.catch_warnings(record=True) as warning_list:
                sts.band(obsmode)
                for warning in warning_list:
                    warns.append("Warning from stsynphot with obsmode '{}': {}".format(obsmode, warning.message))
        except Exception as e:
            errors.append("Exception from stsynphot with obsmode '{}': {}".format(obsmode, repr(e)))
            result = False

    with warnings.catch_warnings(record=True) as warning_list:
        with _configured_pysynphot(synphot_root) as pys:
            try:
                pys.ObsBandpass(obsmode)
            except Exception as e:
                errors.append("Exception from pysynphot with obsmode '{}': {}".format(obsmode, repr(e)))
                result = False
        for warning in warning_list:
            if not str(warning.message).startswith("Extinction files not found in"):
                warns.append("Warning from pysynphot with obsmode '{}': {}".format(obsmode, warning.message))

    return result, errors, warns


class SynphotObsmodeIntegrationTest(SynphotIntegrationTestBase):
    """
    Test synphot files by constructing stsynphot objects from the
    observation mode strings in the obsmode reference file.  Requires
    that stsynphot be installed.
    """
    def __init__(
        self,
        context,
        synphot_root,
        obsmodes_file=None,
        link=False,
        processes=1,
        batch_size=100,
        **kwargs
    ):
        super().__init__(context, **kwargs)

        self.synphot_root = synphot_root

        if obsmodes_file is None:
            self.obsmodes_file = utils.get_cache_path(self.context, utils.OBSMODES_REFTYPE)
        else:
            self.obsmodes_file = obsmodes_file

        self.link = link
        self.processes = processes
        self.batch_size = batch_size

    def populate_synphot_root(self):
        """
        Copy or link the full set of synphot files to synphot_root, with
        newly delivered files overlaid.
        """
        os.makedirs(self.synphot_root, exist_ok=True)
        os.makedirs(utils.get_mtab_path(self.synphot_root), exist_ok=True)
        for instrument in utils.ALL_INSTRUMENTS:
            os.makedirs(utils.get_instrument_path(self.synphot_root, instrument), exist_ok=True)

        self.link_or_copy_table(utils.GRAPH_REFTYPE, self.graph_file)
        self.link_or_copy_table(utils.THROUGHPUT_LOOKUP_REFTYPE, self.throughput_lookup_file)
        self.link_or_copy_table(utils.THERMAL_LOOKUP_REFTYPE, self.thermal_lookup_file)

        for component, path in self.throughput_compname_to_path.items():
            self.link_or_copy_component(component, path)

        for component, path in self.thermal_compname_to_path.items():
            self.link_or_copy_component(component, path)

    def run(self, progress_callback=None):
        """
        Create stsynphot bandpass objects from the observation modes in the obsmodes file.
        Emits appropriate log messages and returns True if validations succeed,
        False if there's an error.
        """

        failed = 0
        with fits.open(self.obsmodes_file) as hdul:
            total_modes = len(hdul[-1].data)
            log.info("Creating bandpass objects from {} observation modes".format(total_modes))

            if self.processes > 1:
                with Pool(processes=self.processes) as pool:
                    for start_index in range(0, total_modes, self.batch_size):
                        end_index = start_index + self.batch_size
                        results = pool.starmap(
                            _test_synphot_mode,
                            [(self.synphot_root, m) for m in hdul[-1].data["OBSMODE"][start_index:end_index]]
                        )
                        for i, (result, errors, warns) in enumerate(results):
                            if not result:
                                failed += 1
                            for warning in warns:
                                log.warning(warning)
                            for error in errors:
                                log.error(error)

                            if progress_callback:
                                progress_callback(start_index + i + 1, total_modes)
            else:
                for i, obsmode in enumerate(hdul[-1].data["OBSMODE"]):
                    result, errors, warns = _test_synphot_mode(self.synphot_root, obsmode)
                    if not result:
                        failed += 1
                    for warning in warns:
                        log.warning(warning)
                    for error in errors:
                        log.error(error)

                    if progress_callback:
                        progress_callback(i + 1, total_modes)

        if failed > 0:
            log.info("{} / {} observation modes failed".format(failed, total_modes))
        else:
            log.info("Congratulations, all observation modes succeeded!")

        return failed == 0

    def link_or_copy_component(self, component, source):
        source_basename = os.path.basename(source)
        instrument = utils.get_instrument(component)
        dest = os.path.join(utils.get_instrument_path(self.synphot_root, instrument), source_basename)
        self.link_or_copy(source, dest)

    def link_or_copy_table(self, reftype, source):
        # Don't use the original filename here, in case a replacement
        # graph or lookup was uploaded with an unusual name.
        if reftype == utils.GRAPH_REFTYPE:
            basename = "test_tmg.fits"
        elif reftype == utils.THROUGHPUT_LOOKUP_REFTYPE:
            basename = "test_tmc.fits"
        else:
            basename = "test_tmt.fits"
        dest = os.path.join(utils.get_mtab_path(self.synphot_root), basename)
        self.link_or_copy(source, dest)

    def link_or_copy(self, source, dest):
        if os.path.exists(dest):
            return

        if self.link:
            os.link(source, dest)
        else:
            shutil.copy(source, dest)


class SynphotCoreIntegrationTest(SynphotIntegrationTestBase):
    """
    Perform additional validations on the full ensemble of synphot files.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self):
        """
        Combine newly delivered files with existing synphot files and validate
        the group for consistency.  Emits appropriate log messages and returns
        True if validations succeed, False if there's an error.
        """
        graph = fits.open(self.graph_file)[-1].data
        graph_throughput_compnames = set(graph["COMPNAME"])
        graph_thermal_compnames = set(graph["THCOMPNAME"])

        throughput_lookup = fits.open(self.throughput_lookup_file)[-1].data
        throughput_lookup_compnames = set(throughput_lookup["COMPNAME"])

        thermal_lookup = fits.open(self.thermal_lookup_file)[-1].data
        thermal_lookup_compnames = set(thermal_lookup["COMPNAME"])

        result = self.check_compname_agreement(
            "TMG COMPNAME",
            graph_throughput_compnames,
            set(),
            "TMC COMPNAME",
            throughput_lookup_compnames,
            set()
        )

        result = self.check_compname_agreement(
            "TMG THCOMPNAME",
            graph_thermal_compnames,
            utils.KNOWN_MISSING_GRAPH_THERMAL_COMPONENTS,
            "TMT COMPNAME",
            thermal_lookup_compnames,
            utils.KNOWN_MISSING_THERMAL_LOOKUP_COMPONENTS
        ) and result

        result = self.check_compname_agreement(
            "TMG COMPNAME",
            graph_throughput_compnames,
            set(),
            "throughput table COMPNAME",
            set(self.throughput_compname_to_path.keys()),
            utils.KNOWN_MISSING_THROUGHPUT_COMPONENTS
        ) and result

        result = self.check_compname_agreement(
            "TMG THCOMPNAME",
            graph_thermal_compnames,
            utils.KNOWN_MISSING_GRAPH_THERMAL_COMPONENTS,
            "thermal table COMPNAME",
            set(self.thermal_compname_to_path.keys()),
            utils.KNOWN_MISSING_THERMAL_COMPONENTS.union(utils.KNOWN_MISSING_THERMAL_LOOKUP_COMPONENTS)
        ) and result

        result = self.check_filenames(
            "TMC",
            throughput_lookup,
            self.throughput_compname_to_path
        ) and result

        result = self.check_filenames(
            "TMT",
            thermal_lookup,
            self.thermal_compname_to_path
        ) and result

        return result

    def check_compname_agreement(
        self,
        description_a,
        compnames_a,
        known_missing_a,
        description_b,
        compnames_b,
        known_missing_b,
    ):
        """
        Check and report any differences between two sets of component names.
        """
        result = True

        log.info("Checking for components present in {} but missing from {}".format(
            description_a, description_b
        ))
        missing_from_b = (compnames_a - compnames_b) - known_missing_b
        if len(missing_from_b) > 0:
            missing_compnames = ", ".join(missing_from_b)
            message = "Components present in {} but missing from {}: {}".format(
                description_a, description_b, missing_compnames
            )
            log.error(message)
            result = False

        log.info("Checking for components present in {} but missing from {}".format(
            description_b, description_a
        ))
        missing_from_a = (compnames_b - compnames_a) - known_missing_a
        if len(missing_from_a) > 0:
            missing_compnames = ", ".join(missing_from_a)
            message = "Components present in {} but missing from {}: {}".format(
                description_b, description_a, missing_compnames
            )
            log.error(message)
            result = False

        return result

    def check_filenames(self, description, lookup, compname_to_path):
        """
        Check that lookup filenames are correct.
        """
        result = True

        log.info("Confirming correctly formed {} filenames".format(description))
        for row in lookup:
            lookup_filename = utils.get_lookup_filename(
                row["COMPNAME"],
                os.path.basename(compname_to_path[row["COMPNAME"]])
            )
            if lookup_filename != row["FILENAME"]:
                log.error(
                    "Malformed {} filename, expected '{}', found '{}'".format(
                        description, lookup_filename, row["FILENAME"]
                    )
                )
                result = False

        return result



"""This module defines limited facilities for extracting information from
reference and datasets,  generally in the form of header dictionaries.
"""
from crds.core  import utils, log

# =============================================================================

from crds.io.abstract import hijack_warnings, convert_to_eval_header, ensure_keys_defined
from crds.io.factory import file_factory, get_observatory, get_filetype, is_dataset
from crds.io.geis import is_geis, is_geis_data, is_geis_header, get_conjugate
from crds.io.fits import fits_open, fits_open_trapped, get_fits_header_union

# import asdf
# import yaml

# =============================================================================

def get_conditioned_header(filepath, needed_keys=(), original_name=None, observatory=None):
    """Return the complete conditioned header dictionary of a reference file,
    or optionally only the keys listed by `needed_keys`.

    DOES NOT verify checksums.

    `original_name`,  if specified,  is used to determine the type of the file
    and is not required to be readable,  whereas `filepath` must be readable
    and contain the desired header.
    """
    header = get_header(filepath, needed_keys, original_name, observatory=observatory)
    return utils.condition_header(header, needed_keys)

@hijack_warnings
def get_header(filepath, needed_keys=(), original_name=None, observatory=None):
    """Return the complete unconditioned header dictionary of a reference file.

    Hijack io.fits and data model warnings and map them to errors.

    DOES NOT verify checksums.

    Original name is used to determine file type for web upload temporary files which
    have no distinguishable extension.  Original name is browser-side name for file.
    """
    return get_free_header(filepath, needed_keys, original_name, observatory)

# A clearer name
get_unconditioned_header = get_header

@utils.cached
# @utils.gc_collected
def get_free_header(filepath, needed_keys=(), original_name=None, observatory=None):
    """Return the complete unconditioned header dictionary of a reference file.

    DOES NOT hijack warnings.   DOES NOT verify checksums.

    Original name is used to determine file type for web upload temporary files
    which have no distinguishable extension.  Original name is browser-side
    name for file.

    get_free_header() is a cached function to prevent repeat file reads.
    Although parameters are given default values,  for caching to work correctly
    even default parameters should be specified positionally.

    Since get_free_header() is cached,  loading file updates requires first
    clearing the function cache.
    """
    file_obj = file_factory(filepath, original_name, observatory)
    header = file_obj.get_header(needed_keys, checksum=False)
    log.verbose("Header of", repr(filepath), "=", log.PP(header), verbosity=90)
    return header

def clear_header_cache():
    """Flush the header cache,  nominally to recover storage taken by array attributes
    brought in for certify.
    """
    get_free_header.cache.clear()

# ================================================================================================================

# @hijack_warnings
def getval(filepath, key, condition=True):
    """Return a single metadata value from `key` of file at `filepath`."""
    if key.upper().startswith("META_"):
        key = key.replace("META_", "META.")
    file_obj = file_factory(filepath)
    value = file_obj.getval(key)
    value = utils.condition_value(value) if condition else value
    return value

@hijack_warnings
@utils.gc_collected
def setval(filepath, key, value):
    """Set metadata keyword `key` of `filepath` to `value`."""
    if key.upper().startswith("META_"):
        key = key.replace("META_", "META.")
    file_obj = file_factory(filepath)
    file_obj.setval(key, value)

def add_checksum(filepath):
    """Add checksums to `filepath`."""
    file_obj = file_factory(filepath)
    return file_obj.add_checksum()

def remove_checksum(filepath):
    """Remove checksums to `filepath`."""
    file_obj = file_factory(filepath)
    return file_obj.remove_checksum()

def verify_checksum(filepath):
    """Verify checksums in `filepath`."""
    file_obj = file_factory(filepath)
    return file_obj.verify_checksum()

def get_asdf_standard_version(filepath):
    """
    Retrieve the ASDF Standard version associated with the file,
    or `None` if not applicable.
    """
    file_obj = file_factory(filepath)
    return file_obj.get_asdf_standard_version()

# ================================================================================================================

def get_array_properties(filename, array_name, keytype="A"):
    """Return the dictionary defining basic properties of `array_name` of `filename`.

    Keytype == "A" for "array" means lightweight format only checks with no data included.
    Keytype == "D" for "data" means heavy data oriented checks with data arrays returned as well.
    """
    file_obj = file_factory(filename)
    props = file_obj.get_array_properties(array_name, keytype)
    return props

# ================================================================================================================

def test():
    """Run doctest on data_file module."""
    import doctest
    from crds import data_file
    return doctest.testmod(data_file)

if __name__ == "__main__":
    print(test())



"""This module supplies the @background decorator and background_complete() results
wait/retrieval function.
"""
import sys
import threading

from crds import exceptions

# ==================================================================================================

class ThreadWithExceptions(threading.Thread):
    """ThreadWithExceptions captures any exception raised during the Thread and stores
    both the result or the exception str() and repr() as attributes.
    """
    def __init__(self, target, *args, **keys):
        super(ThreadWithExceptions, self).__init__(*args, **keys)
        self.target = target
        self.exc = None
        self.rexc = None
        self.result = None

    def run(self):
        """Run the target function, capturing both the return value and any exception
        raised.
        """
        try:
            self.result = self.target()
            self.exc = None
        except:
            _type, exc, _traceb = sys.exc_info()
            self.result = None
            self.rexc = repr(exc)
            self.exc = str(exc)

def background(func):
    """A threading decorator use @background above the function you want to run in
    the background.

    The decorated function returns a daemon thread which allows the main
    program to exit without background task completion.  Unfinished background
    threads shutdown abruptly.
    """

    def run_thread(*args, **keys):
        """Function wrapper which runs the function as a background thread and
        returns the started daemon thread.
        """
        def target_func():
            """Binding function which creates closure of function with passed args
            for purposes of parameterless function for threading.
            """
            return func(*args, **keys)
        thread = ThreadWithExceptions(target=target_func, name=func.__name__)
        thread.daemon = True
        thread.start()
        return thread

    run_thread.__name__ = func.__name__ + "[background]"

    return run_thread

def background_complete(thread):
    """Complete a background task `t` which is either a ThreadWithExceptions object or
    an ordinary return value (which is not a thread).  This structure facilitates un-decorating
    background tasks to make them ordinary functions without restructuring overall control-flow.

    Does a join() iterating every second while thread alive to provide
    foreground thread a guaranteed place to recieve exceptions.

    If background task `t` raises an exception it is reraised here.

    Returns the result of background task.
    """
    if isinstance(thread, threading.Thread):
        while thread.is_alive():
            thread.join(1.0)
        if thread.exc is None:
            return thread.result
        else:
            raise exceptions.CrdsBackgroundError("Exception in", repr(thread.name), ":", repr(thread.exc))
    else:
        return thread

# ==================================================================================================



"""uses.py defines functions which will list the files which use a given
reference or mapping file.

>> from pprint import pprint as pp
>> pp(_findall_mappings_using_reference("v2e20129l_flat.fits"))
['hst.pmap',
 'hst_0001.pmap',
 'hst_0002.pmap',
 'hst_0003.pmap',
 'hst_0004.pmap',
 'hst_0005.pmap',
 'hst_0006.pmap',
 'hst_cos.imap',
 'hst_cos_0001.imap',
 'hst_cos_flatfile.rmap',
 'hst_cos_flatfile_0002.rmap']
"""
import sys
import os.path

from crds.core import config, cmdline, utils, log, rmap

@utils.cached
def load_all_mappings(observatory, pattern="*map"):
    """Return a dictionary mapping the names of all CRDS Mappings matching `pattern`
    onto the loaded Mapping object.
    """
    all_mappings = rmap.list_mappings(pattern, observatory)
    loaded = {}
    for name in all_mappings:
        with log.error_on_exception("Failed loading", repr(name)):
            loaded[name] = rmap.get_cached_mapping(name)
    return loaded

@utils.cached
def mapping_type_names(observatory, ending):
    """Return a mapping dictionary containing only mappings with names with `ending`."""
    return { name : mapping.mapping_names() + mapping.reference_names()
             for (name, mapping) in load_all_mappings(observatory).items()
             if name.endswith(ending) }

def uses_files(files, observatory, ending):
    """Alternate approach to uses that works by loading all mappings instead of grepping
    them on the file system.
    """
    referrers = set()
    for filename in files:
        config.check_filename(filename)
        loaded = mapping_type_names(observatory, ending)
        for_filename = set(name for name in loaded if filename in loaded[name])
        referrers |= for_filename
    return sorted(list(referrers))

def _findall_rmaps_using_reference(filename, observatory="hst"):
    """Return the basename of all reference mappings which mention `filename`."""
    return uses_files([filename], observatory, "rmap")

def _findall_imaps_using_rmap(filename, observatory="hst"):
    """Return the basenames of all instrument contexts which mention `filename`."""
    return uses_files([filename], observatory, "imap")

def _findall_pmaps_using_imap(filename, observatory="hst"):
    """Return the basenames of all pipeline contexts which mention `filename`."""
    return uses_files([filename], observatory, "pmap")

def _findall_mappings_using_reference(reference, observatory="hst"):
    """Return the basenames of all mapping files in the hierarchy which
    mentions reference `reference`.
    """
    mappings = []
    for rmap in _findall_rmaps_using_reference(reference, observatory):
        mappings.append(rmap)
        for imap in _findall_imaps_using_rmap(rmap, observatory):
            mappings.append(imap)
            for pmap in _findall_pmaps_using_imap(imap, observatory):
                mappings.append(pmap)
    return sorted(list(set(mappings)))

def _findall_mappings_using_rmap(rmap, observatory="hst"):
    """Return the basenames of all mapping files in the hierarchy which
    mentions reference mapping `rmap`.
    """
    mappings = []
    for imap in _findall_imaps_using_rmap(rmap, observatory):
        mappings.append(imap)
        for pmap in _findall_pmaps_using_imap(imap, observatory):
            mappings.append(pmap)
    return sorted(list(set(mappings)))

def uses(files, observatory="hst"):
    """Return the list of mappings which use any of `files`."""
    mappings = []
    for file_ in files:
        if file_.endswith(".rmap"):
            mappings.extend(_findall_mappings_using_rmap(file_, observatory))
        elif file_.endswith(".imap"):
            mappings.extend(_findall_pmaps_using_imap(file_, observatory))
        elif file_.endswith(".pmap"):
            pass  # nothing refers to a .pmap
        else:
            mappings.extend(_findall_mappings_using_reference(file_, observatory))
    return sorted(list(set(mappings)))

class UsesScript(cmdline.Script):
    """Command line script for printing rmaps using references,  or datasets using references."""

    description = """
Prints out the mappings which refer to the specified mappings or references.

Prints out the datasets which historically used a particular reference as defined by DADSOPS.

IMPORTANT:
   1. You must specify references or rules on which to operate with --files.
   2. You must set CRDS_PATH and CRDS_SERVER_URL to give crds.uses access to CRDS mappings and databases.
"""

    epilog = """
crds.uses can be invoked like this:

% crds uses --files n3o1022ij_drk.fits --hst
hst.pmap
hst_0001.pmap
hst_0002.pmap
hst_0003.pmap
...
hst_0041.pmap
hst_acs.imap
hst_acs_0001.imap
hst_acs_0002.imap
hst_acs_0003.imap
...
hst_acs_0008.imap
hst_acs_darkfile.rmap
hst_acs_darkfile_0001.rmap
hst_acs_darkfile_0002.rmap
hst_acs_darkfile_0003.rmap
...
hst_acs_darkfile_0005.rmap

"""
    def add_args(self):
        """Add command line parameters unique to this script."""
        super(UsesScript, self).add_args()
        self.add_argument("--files", nargs="+",
                          help="References for which to dump using mappings or datasets.")
        self.add_argument("-i", "--include-used", action="store_true", dest="include_used",
                          help="Include the used file in the output as the first column.")

    def main(self):
        """Process command line parameters in to a context and list of
        reference files.   Print out the match tuples within the context
        which contain the reference files.
        """
        if not self.args.files:
            self.print_help()
            sys.exit(-1)
        self.print_mappings_using_files()
        return log.errors()

    def locate_file(self, file_):
        """Just use basenames for identifying file references."""
        return os.path.basename(file_)

    def print_mappings_using_files(self):
        """Print out the mappings which refer to the specified mappings or references."""
        for file_ in self.files:
            for use in uses([file_], self.observatory):
                if self.args.include_used:
                    print(file_, use)
                else:
                    print(use)

def test():
    """Run the module doctest."""
    import doctest
    from . import uses
    return doctest.testmod(uses)

if __name__ == "__main__":
    sys.exit(UsesScript()())



"""This module is a command line script which lists the match tuples associated
with a reference file.

% crds matches  --contexts hst_0001.pmap --files lc41311jj_pfl.fits
lc41311jj_pfl.fits : ACS PFLTFILE DETECTOR='WFC' CCDAMP='A|ABCD|AC|AD|B|BC|BD|C|D' FILTER1='F625W' FILTER2='POL0V' OBSTYPE='IMAGING' FW1OFFST='N/A' FW2OFFST='N/A' FWSOFFST='N/A' DATE-OBS='1997-01-01' TIME-OBS='00:00:00'

A number of command line switches control output formatting.

The api function find_full_match_paths() returns a list of "match paths",  lists of parkey value assignment tuples:
"""
import sys
import os.path
from collections import defaultdict
from pprint import pprint as pp   # doctests

import crds
from crds.core import log, utils, cmdline, selectors
from crds.client import api

# ===================================================================

def test():
    """Run any doctests."""
    import doctest, crds.matches
    return doctest.testmod(crds.matches)

# ===================================================================

def find_full_match_paths(context, reffile):
    """Return the list of full match paths for `reference` in `context` as a
    list of tuples of tuples.   Each inner tuple is a (var, value) pair.

    Returns [((context_tuples,),(match_tuple,),(useafter_tuple,)), ...]

    >>> pp(find_full_match_paths("hst.pmap", "q9e1206kj_bia.fits"))
    [((('observatory', 'hst'), ('instrument', 'acs'), ('filekind', 'biasfile')),
      (('DETECTOR', 'HRC'),
       ('CCDAMP', 'A'),
       ('CCDGAIN', '4.0'),
       ('APERTURE', '*'),
       ('NUMCOLS', '1062'),
       ('NUMROWS', '1044'),
       ('LTV1', '19.0'),
       ('LTV2', '20.0'),
       ('XCORNER', 'N/A'),
       ('YCORNER', 'N/A'),
       ('CCDCHIP', 'N/A')),
      (('DATE-OBS', '2006-07-04'), ('TIME-OBS', '11:32:35'))),
     ((('observatory', 'hst'), ('instrument', 'acs'), ('filekind', 'biasfile')),
      (('DETECTOR', 'HRC'),
       ('CCDAMP', 'A'),
       ('CCDGAIN', '4.0'),
       ('APERTURE', '*'),
       ('NUMCOLS', 'N/A'),
       ('NUMROWS', 'N/A'),
       ('LTV1', 'N/A'),
       ('LTV2', 'N/A'),
       ('XCORNER', 'N/A'),
       ('YCORNER', 'N/A'),
       ('CCDCHIP', 'N/A')),
      (('DATE-OBS', '2006-07-04'), ('TIME-OBS', '11:32:35')))]
    """
    ctx = crds.get_pickled_mapping(context, cached=True)  # reviewed
    return ctx.file_matches(reffile)

def find_match_paths_as_dict(context, reffile):
    """Return the matching parameters for reffile as a list of dictionaries, one dict for
    each match case giving the parameters of that match.

    >>> pp(find_match_paths_as_dict("hst.pmap", "q9e1206kj_bia.fits"))
    [{'APERTURE': '*',
      'CCDAMP': 'A',
      'CCDCHIP': 'N/A',
      'CCDGAIN': '4.0',
      'DATE-OBS': '2006-07-04',
      'DETECTOR': 'HRC',
      'LTV1': '19.0',
      'LTV2': '20.0',
      'NUMCOLS': '1062',
      'NUMROWS': '1044',
      'TIME-OBS': '11:32:35',
      'XCORNER': 'N/A',
      'YCORNER': 'N/A',
      'filekind': 'biasfile',
      'instrument': 'acs',
      'observatory': 'hst'},
     {'APERTURE': '*',
      'CCDAMP': 'A',
      'CCDCHIP': 'N/A',
      'CCDGAIN': '4.0',
      'DATE-OBS': '2006-07-04',
      'DETECTOR': 'HRC',
      'LTV1': 'N/A',
      'LTV2': 'N/A',
      'NUMCOLS': 'N/A',
      'NUMROWS': 'N/A',
      'TIME-OBS': '11:32:35',
      'XCORNER': 'N/A',
      'YCORNER': 'N/A',
      'filekind': 'biasfile',
      'instrument': 'acs',
      'observatory': 'hst'}]
    """
    matches = find_full_match_paths(context, reffile)
    return [ _flatten_items_to_dict(match) for match in matches ]

def _flatten_items_to_dict(match_path):
    """Given a `match_path` which is a sequence of parameter items and sub-paths,  return
    a flat dictionary representation:

    returns   { matchinhg_par:  matching_par_value, ...}
    """
    result = {}
    for par in match_path:
        if len(par) == 2 and isinstance(par[0], str) and isinstance(par[1], str):
            result[par[0]] = par[1]
        else:
            result.update(_flatten_items_to_dict(par))
    return result

def get_minimum_exptime(context, references):
    """Return the minimum EXPTIME for the list of `references` with respect to `context`.
    This is used to define the potential reprocessing impact of new references,  since
    no dataset with an earlier EXPTIME than a reference is typically affected by the
    reference,  particularly with respect to the HST USEAFTER approach.

    >>> get_minimum_exptime("hst.pmap", ["q9e1206kj_bia.fits"])
    '2006-07-04 11:32:35'
    """
    return min([_get_minimum_exptime(context, ref) for ref in references])

def _get_minimum_exptime(context, reffile):
    """Given a `context` and a `reffile` in it,  return the minimum EXPTIME for all of
    it's match paths constructed from DATE-OBS and TIME-OBS.
    """
    path_dicts = find_match_paths_as_dict(context, reffile)
    exptimes = [ get_exptime(path_dict) for path_dict in path_dicts ]
    return min(exptimes)


DATE_TIME_PAIRS = [
    ("DATE-OBS", "TIME-OBS"),
    ("META.OBSERVATION.DATE", "META.OBSERVATION.TIME"),
    ("META_OBSERVATION_DATE", "META_OBSERVATION_TIME"),
    ]

def get_exptime(match_dict):
    """Given a `match_dict` dictionary of matching parameters for one match path,
    return the EXPTIME for it or 1900-01-01 00:00:00 if no EXPTIME can be derived.
    """
    for dt_pair in DATE_TIME_PAIRS:
        try:
            return match_dict[dt_pair[0]] + " " + match_dict[dt_pair[1]]
        except KeyError:
            continue
    else:
        log.verbose("matches.exp_time:  no exptime value found. returning worst case 1900-01-01 00:00:00.")
        return "1900-01-01 00:00:00"

# ===================================================================

class MatchesScript(cmdline.ContextsScript):
    """Command line script for printing reference selection criteria."""

    description = """
Prints out the selection criteria by which the specified references are matched with respect to the specified contexts.

The primary and original role of crds.matches is to interpret CRDS rules and report the matching criteria for specified
references.

A secondary function of crds.matches is to dump the matching criteria associated with particular dataset ids,
or all dataset ids for an instrument, according to the appropriate archive catalog (e.g. DADSOPS).
"""

    epilog = """
** crds.matches can dump reference file match cases with respect to particular contexts:

% crds matches  --contexts hst_0001.pmap --files lc41311jj_pfl.fits
lc41311jj_pfl.fits : ACS PFLTFILE DETECTOR='WFC' CCDAMP='A|ABCD|AC|AD|B|BC|BD|C|D' FILTER1='F625W' FILTER2='POL0V' DATE-OBS='1997-01-01' TIME-OBS='00:00:00'

% crds matches --contexts hst.pmap --files lc41311jj_pfl.fits --omit-parameter-names --brief-paths
lc41311jj_pfl.fits :  'WFC' 'A|ABCD|AC|AD|B|BC|BD|C|D' 'F625W' 'POL0V' '1997-01-01' '00:00:00'

% crds matches --contexts hst.pmap --files lc41311jj_pfl.fits --tuple-format
lc41311jj_pfl.fits : (('OBSERVATORY', 'HST'), ('INSTRUMENT', 'ACS'), ('FILEKIND', 'PFLTFILE'), ('DETECTOR', 'WFC'), ('CCDAMP', 'A|ABCD|AC|AD|B|BC|BD|C|D'), ('FILTER1', 'F625W'), ('FILTER2', 'POL0V'), ('DATE-OBS', '1997-01-01'), ('TIME-OBS', '00:00:00'))

** crds.matches can dump database matching parameters for specified datasets with respect to specified contexts:

% crds matches --datasets JBANJOF3Q --minimize-headers --contexts hst_0048.pmap hst_0044.pmap
JBANJOF3Q : hst_0044.pmap : APERTURE='WFC1-2K' ATODCORR='NONE' BIASCORR='NONE' CCDAMP='B' CCDCHIP='1.0' CCDGAIN='2.0' CRCORR='NONE' DARKCORR='NONE' DATE-OBS='2010-01-31' DETECTOR='WFC' DQICORR='NONE' DRIZCORR='NONE' FILTER1='F502N' FILTER2='F660N' FLASHCUR='OFF' FLATCORR='NONE' FLSHCORR='NONE' FW1OFFST='0.0' FW2OFFST='0.0' FWSOFFST='0.0' GLINCORR='NONE' INSTRUME='ACS' LTV1='-2048.0' LTV2='-1.0' NUMCOLS='UNDEFINED' NUMROWS='UNDEFINED' OBSTYPE='INTERNAL' PCTECORR='NONE' PHOTCORR='NONE' REFTYPE='UNDEFINED' SHADCORR='NONE' SHUTRPOS='B' TIME-OBS='01:07:14.960000' XCORNER='1.0' YCORNER='2072.0'
JBANJOF3Q : hst_0048.pmap : APERTURE='WFC1-2K' ATODCORR='NONE' BIASCORR='NONE' CCDAMP='B' CCDCHIP='1.0' CCDGAIN='2.0' CRCORR='NONE' DARKCORR='NONE' DATE-OBS='2010-01-31' DETECTOR='WFC' DQICORR='NONE' DRIZCORR='NONE' FILTER1='F502N' FILTER2='F660N' FLASHCUR='OFF' FLATCORR='NONE' FLSHCORR='NONE' FW1OFFST='0.0' FW2OFFST='0.0' FWSOFFST='0.0' GLINCORR='NONE' INSTRUME='ACS' LTV1='-2048.0' LTV2='-1.0' NAXIS1='2070.0' NAXIS2='2046.0' OBSTYPE='INTERNAL' PCTECORR='NONE' PHOTCORR='NONE' REFTYPE='UNDEFINED' SHADCORR='NONE' SHUTRPOS='B' TIME-OBS='01:07:14.960000' XCORNER='1.0' YCORNER='2072.0'

** crds.matches can list all references which satisfy any filter constraints relevant to their bestref lookup.

% crds matches --contexts jwst-niriss-operational --files-from-contexts --filters META.INSTRUMENT.FILTER='F480M' --brief
CRDS - INFO - Symbolic context 'jwst-niriss-operational' resolves to 'jwst_niriss_0028.imap'
jwst_niriss_dark_0005.fits :  META.INSTRUMENT.DETECTOR='NIS' META.SUBARRAY.NAME='FULL'
jwst_niriss_gain_0001.fits :  META.INSTRUMENT.DETECTOR='NIS'
jwst_niriss_ipc_0002.fits :  META.INSTRUMENT.DETECTOR='NIS'
jwst_niriss_linearity_0005.fits :  META.INSTRUMENT.DETECTOR='NIS'
jwst_niriss_mask_0004.fits :  META.INSTRUMENT.DETECTOR='NIS'
jwst_niriss_photom_0017.fits :  META.INSTRUMENT.DETECTOR='NIS'
jwst_niriss_readnoise_0001.fits :  META.INSTRUMENT.DETECTOR='NIS'
jwst_niriss_saturation_0005.fits :  META.INSTRUMENT.DETECTOR='NIS'
jwst_niriss_superbias_0003.fits :  META.INSTRUMENT.DETECTOR='NIS' META.EXPOSURE.READPATT='*' META.SUBARRAY.NAME='N/A'
jwst_niriss_throughput_0008.fits :  META.INSTRUMENT.FILTER='F480M'
"""

    def add_args(self):
        super(MatchesScript, self).add_args()
        self.add_argument("--files", nargs="+", default=(),
            help="References for which to dump selection criteria.")
        self.add_argument("--files-from-contexts", action="store_true",
            help="Operate on all references referred to by the context parameters.")
        self.add_argument("--filters", nargs="+", default=(),
            help="Parameter constraints (key=value) which references matching on `key` must satisfy.  Unused parameters for a reference type are ignored.")
        self.add_argument("-b", "--brief-paths", action="store_true",
            help="Don't show the instrument and filekind clutter if already in filename.")
        self.add_argument("-o", "--omit-parameter-names", action="store_true",
            help="Hide the parameter names of the selection criteria,  just show the values.")
        self.add_argument("-t", "--tuple-format", action="store_true",
            help="Print the match info as Python tuples.")
        self.add_argument("-d", "--datasets", nargs="+",
            help="Dataset ids for which to dump matching parameters from DADSOPS or equivalent database.")
        self.add_argument("-i", "--instrument", type=str,
            help="Instrument for which to dump matching parameters from DADSOPS or equivalent database.")
        self.add_argument("-c", "--condition-values", action="store_true",
            help="When dumping dataset parameters, first apply CRDS value conditioning / normalization.")
        self.add_argument("-m", "--minimize-headers", action="store_true",
            help="When dumping dataset parameters,  limit them to matching parameters, excluding e.g. historical bestrefs.")

    def main(self):
        """Process command line parameters in to a context and list of
        reference files.   Print out the match tuples within the context
        which contain the reference files.
        """
        if self.matched_files:
            self.dump_reference_matches()
        elif self.args.datasets or self.args.instrument:
            self.dump_dataset_headers()
        else:
            self.print_help()
            log.error("Specify --files to dump reference match cases or --datasets to dump dataset matching parameters.")
        return log.errors()

    @property
    def matched_files(self):
        """Combine references from --files with references implied by --contexts parameters."""
        matched = list(self.files)
        if self.args.files_from_contexts:
            matched += list(self.get_context_references())
        return matched

    def dump_reference_matches(self):
        """Print out the match paths for the reference files specified on the
        command line with respect to the specified contexts.
        """
        for ref in self.matched_files:
            cmdline.reference_file(ref)
        for context in self.contexts:
            self.dump_match_tuples(context)

    def dump_dataset_headers(self):
        """Print out the matching parameters for the --datasets specified on
        the command line.
        """
        multi_context_headers = defaultdict(list)
        for context in self.contexts:
            if self.args.datasets:
                headers = api.get_dataset_headers_by_id(context, self.args.datasets)
            elif self.args.instrument:
                headers = api.get_dataset_headers_by_instrument(context, self.args.instrument)
            for dataset_id, header in headers.items():
                multi_context_headers[dataset_id].append((context, header))
        for dataset_id, context_headers in multi_context_headers.items():
            for (context, header) in context_headers:
                if self.args.condition_values:
                    header = utils.condition_header(header)
                if self.args.minimize_headers:
                    header = crds.get_cached_mapping(context).minimize_header(header)
                if len(self.contexts) == 1:
                    print(dataset_id, ":", log.format_parameter_list(header))
                else:
                    print(dataset_id, ":", context, ":", log.format_parameter_list(header))

    def locate_file(self, filename):
        """Override for self.files..."""
        return os.path.basename(filename)

    def dump_match_tuples(self, context):
        """Print out the match tuples for `references` under `context`.
        """
        ctx = context if len(self.contexts) > 1 else ""
        for ref in self.matched_files:
            matches = self.find_match_tuples(context, ref)
            if matches:
                for match in matches:
                    log.write(ctx, ref, ":", match)
            else:
                log.verbose(ctx, ref, ":", "none")

    def find_match_tuples(self, context, reffile):
        """Return the list of match representations for `reference` in `context`.
        """
        ctx = crds.get_cached_mapping(context)
        matches = ctx.file_matches(reffile)
        result = []
        for path in matches:
            prefix = self.format_prefix(path[0])
            if self.is_filtered(path):
                continue
            match_tuple = tuple([self.format_match_tup(tup) for section in path[1:] for tup in section])
            if self.args.tuple_format:
                if prefix:
                    match_tuple = prefix + match_tuple
            else:
                match_tuple = prefix + " " + " ".join(match_tuple)
            result.append(match_tuple)
        return result

    def is_filtered(self, path):
        """Return True is `path` meets all matching parameter constraints specified by --filters,
        otherwise False.
        """
        for filter in self.args.filters:
            key, value = (item.strip() for item in filter.split("="))
            for section in path[1:]:
                for tup in section:
                    if (tup[0].upper() == key.upper() and
                        value.upper() not in selectors.glob_list(tup[1].upper())):
                        return True
        return False

    def format_prefix(self, path):
        """Return any representation of observatory, instrument, and filekind."""
        if not self.args.brief_paths:
            if self.args.tuple_format:
                prefix = tuple([tuple([t.upper() for t in tup]) for tup in path])
            else:
                prefix = " ".join(tup[1].upper() for tup in path[1:])
        else:
            prefix = ""
        return prefix

    def format_match_tup(self, tup):
        """Return the representation of the selection criteria."""
        if self.args.tuple_format:
            return tup if not self.args.omit_parameter_names else tup[1]
        else:
            tup = tup[0], repr(tup[1])
            return "=".join(tup if not self.args.omit_parameter_names else tup[1:])

if __name__ == "__main__":
   sys.exit(MatchesScript()())
